# Generated by Django 2.0.6 on 2018-11-19 11:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("vng_api_common", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="APICredential",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("api_root", models.URLField(unique=True, verbose_name="api root")),
                (
                    "client_id",
                    models.CharField(max_length=255, verbose_name="client id"),
                ),
                ("secret", models.CharField(max_length=255, verbose_name="secret")),
            ],
            options={
                "verbose_name": "API credential",
                "verbose_name_plural": "API credentials",
            },
        )
    ]



# Generated by Django 2.0.13 on 2019-04-29 14:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("notifications", "0006_auto_20190417_1142")]

    operations = [
        migrations.AlterField(
            model_name="notificationsconfig",
            name="api_root",
            field=models.URLField(unique=True, verbose_name="api root"),
        )
    ]



# Generated by Django 2.1.1 on 2019-03-19 09:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("notifications", "0002_subscription__subscription")]

    operations = [
        migrations.RenameField(
            model_name="notificationsconfig", old_name="location", new_name="api_root"
        ),
        migrations.AddField(
            model_name="notificationsconfig",
            name="client_id",
            field=models.CharField(
                blank=True, max_length=255, verbose_name="client id"
            ),
        ),
        migrations.AddField(
            model_name="notificationsconfig",
            name="secret",
            field=models.CharField(blank=True, max_length=255, verbose_name="secret"),
        ),
        migrations.AlterField(
            model_name="notificationsconfig",
            name="api_root",
            field=models.URLField(unique=True, verbose_name="api root"),
        ),
    ]



# Generated by Django 2.0.13 on 2019-05-16 08:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("audittrails", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="audittrail",
            name="applicatie_id",
            field=models.CharField(
                blank=True,
                help_text="Unieke identificatie van de applicatie, binnen de organisatie",
                max_length=100,
            ),
        ),
        migrations.AddField(
            model_name="audittrail",
            name="applicatie_weergave",
            field=models.CharField(
                blank=True,
                help_text="Vriendelijke naam van de applicatie",
                max_length=200,
            ),
        ),
    ]



import inspect

from django.conf import settings
from django.db import models
from django.urls import reverse as _reverse, reverse_lazy as _reverse_lazy


def _magic_args(args, kwargs):
    """
    Do some trivial introspection to translate objects/models into common
    used urls.
    """
    if args and isinstance(args[0], models.Model):
        url_name = f"{args[0]._meta.model_name}-detail"
        kwargs["kwargs"].setdefault("uuid", args[0].uuid)
        args = (url_name,) + args[1:]
    elif args and inspect.isclass(args[0]) and issubclass(args[0], models.Model):
        url_name = f"{args[0]._meta.model_name}-list"
        args = (url_name,) + args[1:]
    return args, kwargs


def reverse(*args, **kwargs):
    """
    Determine the path for a named URL, instance or model.

    There's some 'magic' behaviour depending on the first arg provided:

    * if a model instance is provided, a detail URL is reversed
    * if a model class is provided, a list URL is reversed

    It is assumed that objects have a ``uuid`` field and the url kwarg is named
    ``uuid``. For nested routes, you need to provide the remaining parameters
    as kwargs. Other Django ``reverse`` params are passed down.
    """
    kwargs.setdefault("kwargs", {})
    kwargs["kwargs"]["version"] = settings.REST_FRAMEWORK["DEFAULT_VERSION"]
    args, kwargs = _magic_args(args, kwargs)
    return _reverse(*args, **kwargs)


def reverse_lazy(*args, **kwargs):
    """
    Determine the path for a named URL, instance or model, lazily.

    There's some 'magic' behaviour depending on the first arg provided:

    * if a model instance is provided, a detail URL is reversed
    * if a model class is provided, a list URL is reversed

    It is assumed that objects have a ``uuid`` field and the url kwarg is named
    ``uuid``. For nested routes, you need to provide the remaining parameters
    as kwargs. Other Django ``reverse_lazy`` params are passed down.
    """
    kwargs.setdefault("kwargs", {})
    kwargs["kwargs"]["version"] = settings.REST_FRAMEWORK["DEFAULT_VERSION"]
    args, kwargs = _magic_args(args, kwargs)
    return _reverse_lazy(*args, **kwargs)



# Generated by Django 2.0.13 on 2019-05-22 09:16

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("audittrails", "0006_audittrail_toelichting")]

    operations = [
        migrations.AlterField(
            model_name="audittrail",
            name="toelichting",
            field=models.TextField(
                blank=True,
                help_text="Toelichting waarom de handeling is uitgevoerd",
                verbose_name="toelichting",
            ),
        )
    ]



import logging
import os
from collections import OrderedDict

from django.apps import apps
from django.conf import settings
from django.http import Http404, HttpRequest
from django.utils.translation import ugettext_lazy as _
from django.views.generic import TemplateView

import requests
from rest_framework import exceptions as drf_exceptions, status
from rest_framework.response import Response
from rest_framework.views import exception_handler as drf_exception_handler
from zds_client import ClientError

from . import exceptions
from .compat import sentry_client
from .constants import ComponentTypes
from .exception_handling import HandledException
from .scopes import SCOPE_REGISTRY

logger = logging.getLogger(__name__)

ERROR_CONTENT_TYPE = "application/problem+json"


def exception_handler(exc, context):
    """
    Transform 4xx and 5xx errors into DSO-compliant shape.
    """
    response = drf_exception_handler(exc, context)
    if response is None:
        if os.getenv("DEBUG", "").lower() in ["yes", "1", "true"]:
            return None

        logger.exception(exc.args[0], exc_info=1)
        # make sure the exception still ends up in Sentry
        sentry_client.captureException()

        # unkown type, so we use the generic Internal Server Error
        exc = drf_exceptions.APIException("Internal Server Error")
        response = Response(status=status.HTTP_500_INTERNAL_SERVER_ERROR)

    request = context.get("request")

    serializer = HandledException.as_serializer(exc, response, request)
    response.data = OrderedDict(serializer.data.items())
    # custom content type
    response["Content-Type"] = ERROR_CONTENT_TYPE
    return response


class ErrorDetailView(TemplateView):
    template_name = "vng_api_common/ref/error_detail.html"

    def _get_exception_klass(self):
        klass = self.kwargs["exception_class"]

        for module in [exceptions, drf_exceptions]:
            exc_klass = getattr(module, klass, None)
            if exc_klass is not None:
                return exc_klass
        else:
            raise Http404("Unknown exception class '{}'".format(klass))

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        exc_klass = self._get_exception_klass()
        context.update(
            {
                "type": exc_klass.__name__,
                "status_code": exc_klass.status_code,
                "default_detail": exc_klass.default_detail,
                "default_code": exc_klass.default_code,
            }
        )
        return context


class ScopesView(TemplateView):
    template_name = "vng_api_common/ref/scopes.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["scopes"] = sorted(
            (scope for scope in SCOPE_REGISTRY if not scope.children),
            key=lambda s: s.label,
        )
        return context


class ViewConfigView(TemplateView):
    template_name = "vng_api_common/view_config.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)

        config = []
        config += _test_sites_config(self.request)
        config += _test_ac_config()
        config += _test_nrc_config()

        context["config"] = config

        return context


def _test_sites_config(request: HttpRequest) -> list:
    if not apps.is_installed("django.contrib.sites"):
        return []

    from django.contrib.sites.models import Site

    site = Site.objects.get_current()

    return [
        (_("Site domain"), site.domain, request.get_host() == site.domain),
        (_("HTTPS"), settings.IS_HTTPS, settings.IS_HTTPS == request.is_secure()),
    ]


def _test_ac_config() -> list:
    if not apps.is_installed("vng_api_common.authorizations"):
        return []

    from .authorizations.models import AuthorizationsConfig

    auth_config = AuthorizationsConfig.get_solo()

    # check if AC auth is configured
    ac_client = AuthorizationsConfig.get_client()
    has_ac_auth = ac_client.auth is not None

    checks = [
        (_("Type of component"), auth_config.get_component_display(), None),
        (_("AC"), auth_config.api_root, auth_config.api_root.endswith("/")),
        (
            _("Credentials for AC"),
            _("Configured") if has_ac_auth else _("Missing"),
            has_ac_auth,
        ),
    ]

    # check if permissions in AC are fine
    if has_ac_auth:
        error = False

        try:
            ac_client.list(
                "applicatie", query_params={"clientIds": ac_client.auth.client_id}
            )
        except requests.ConnectionError:
            error = True
            message = _("Could not connect with AC")
        except ClientError as exc:
            error = True
            message = _(
                "Cannot retrieve authorizations: HTTP {status_code} - {error_code}"
            ).format(status_code=exc.args[0]["status"], error_code=exc.args[0]["code"])
        else:
            message = _("Can retrieve authorizations")

        checks.append((_("AC connection and authorizations"), message, not error))

    return checks


def _test_nrc_config() -> list:
    if not apps.is_installed("vng_api_common.notifications"):
        return []

    from .notifications.models import NotificationsConfig, Subscription

    nrc_config = NotificationsConfig.get_solo()

    nrc_client = NotificationsConfig.get_client()

    has_nrc_auth = nrc_client.auth is not None

    checks = [
        (_("NRC"), nrc_config.api_root, nrc_config.api_root.endswith("/")),
        (
            _("Credentials for NRC"),
            _("Configured") if has_nrc_auth else _("Missing"),
            has_nrc_auth,
        ),
    ]

    # check if permissions in AC are fine
    if has_nrc_auth:
        error = False

        try:
            nrc_client.list("kanaal")
        except requests.ConnectionError:
            error = True
            message = _("Could not connect with NRC")
        except ClientError as exc:
            error = True
            message = _(
                "Cannot retrieve kanalen: HTTP {status_code} - {error_code}"
            ).format(status_code=exc.args[0]["status"], error_code=exc.args[0]["code"])
        else:
            message = _("Can retrieve kanalen")

        checks.append((_("NRC connection and authorizations"), message, not error))

    #  check if there's a subscription for AC notifications
    has_sub = (
        Subscription.objects.filter(channels__contains=["autorisaties"])
        .exclude(_subscription="")
        .exists()
    )
    check_ok = has_sub

    if apps.is_installed("vng_api_common.authorizations"):
        from .authorizations.models import AuthorizationsConfig

        auth_config = AuthorizationsConfig.get_solo()
        if auth_config.component == ComponentTypes.ac and not has_sub:
            check_ok = True

    checks.append(
        (_("Listens to AC notifications?"), _("Yes") if has_sub else _("No"), check_ok)
    )

    return checks



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from .models import APICredential, JWTSecret


@admin.register(JWTSecret)
class JWTSecretAdmin(admin.ModelAdmin):
    list_display = ("identifier",)
    search_fields = ("identifier",)


@admin.register(APICredential)
class APICredentialAdmin(admin.ModelAdmin):
    list_display = ("label", "api_root", "client_id", "user_id")
    search_fields = ("label", "api_root")
    fieldsets = (
        (_("external API"), {"fields": ["api_root", "label"]}),
        (
            _("credentials"),
            {
                "description": _(
                    "Credentials that indicate how this API or application identifies itself at the external "
                    "API."
                ),
                "fields": ["client_id", "secret", "user_id", "user_representation"],
            },
        ),
    )



# Generated by Django 2.0.13 on 2019-05-13 09:31

import uuid

import django.contrib.postgres.fields.jsonb
import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="AuditTrail",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "uuid",
                    models.UUIDField(
                        default=uuid.uuid4,
                        help_text="Unieke resource identifier (UUID4)",
                        unique=True,
                    ),
                ),
                ("bron", models.CharField(max_length=50)),
                ("actie", models.CharField(max_length=50)),
                ("actie_weergave", models.CharField(blank=True, max_length=200)),
                ("resultaat", models.IntegerField()),
                ("hoofd_object", models.URLField(max_length=1000)),
                ("resource", models.CharField(max_length=50)),
                ("resource_url", models.URLField(max_length=1000)),
                ("aanmaakdatum", models.DateTimeField(auto_now=True)),
                (
                    "oud",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                        null=True,
                    ),
                ),
                (
                    "nieuw",
                    django.contrib.postgres.fields.jsonb.JSONField(
                        encoder=django.core.serializers.json.DjangoJSONEncoder,
                        null=True,
                    ),
                ),
            ],
        )
    ]



import re
from typing import Any

from django.core.checks import Warning, register

from djchoices import DjangoChoices

from .utils import get_subclasses

ENUM_VALUE_PATTERN = re.compile(r"^[a-z_0-9]+$", re.ASCII)


def enum_value_ok(value: Any) -> bool:
    if not isinstance(value, str):
        return True

    match = ENUM_VALUE_PATTERN.match(value)
    if not match:
        return False

    return True


@register()
def check_lowercased_constants(app_configs, **kwargs):
    """
    Check that enum values have lowercased, underscore only values.
    """
    warnings = []

    for klass in get_subclasses(DjangoChoices):
        enum_values = klass.values.keys()
        if any((not enum_value_ok(value) for value in enum_values)):
            warnings.append(
                Warning(
                    "Choices %s.%s has at least one value that is not lowercased/underscore ascii only"
                    % (klass.__module__, klass.__name__),
                    hint="Lower case the values and replace dashes/spaces with underscores",
                    obj=klass,
                    id="vng_api_common.enums.W001",
                )
            )

    return warnings



from django.test import override_settings
from django.urls import reverse_lazy

from rest_framework import status
from rest_framework.test import APITestCase

from vng_api_common.tests import JWTAuthMixin

from ..constants import SCOPE_NOTIFICATIES_PUBLICEREN_LABEL


@override_settings(ROOT_URLCONF="vng_api_common.notifications.tests.urls")
class WebhookTests(JWTAuthMixin, APITestCase):

    scopes = [SCOPE_NOTIFICATIES_PUBLICEREN_LABEL]
    url = reverse_lazy("notificaties-webhook")

    def test_auth_required(self):
        self.client.credentials()  # clear any credentials

        response = self.client.post(self.url)

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_receive_nofication(self):
        data = {
            "kanaal": "zaken",
            "hoofdObject": "https://zaken-api.vng.cloud/api/v1/zaken/d7a22",
            "resource": "status",
            "resourceUrl": "https://zaken-api.vng.cloud/api/v1/statussen/d7a22/721c9",
            "actie": "create",
            "aanmaakdatum": "2018-01-01T17:00:00Z",
            "kenmerken": {
                "bron": "082096752011",
                "zaaktype": "https://example.com/api/v1/zaaktypen/5aa5c",
                "vertrouwelijkeidaanduiding": "openbaar",
            },
        }

        response = self.client.post(self.url, data)

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)



# Generated by Django 2.0.13 on 2019-03-12 10:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("notifications", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="subscription",
            name="_subscription",
            field=models.URLField(
                blank=True,
                editable=False,
                help_text="Subscription as it is known in the NC",
                verbose_name="NC subscription",
            ),
        )
    ]



import uuid
from urllib.parse import urljoin

from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.utils.translation import ugettext_lazy as _

from zds_client import Client, ClientAuth

from ..decorators import field_default
from ..models import APICredential, ClientConfig
from .constants import (
    SCOPE_NOTIFICATIES_CONSUMEREN_LABEL,
    SCOPE_NOTIFICATIES_PUBLICEREN_LABEL,
)


@field_default("api_root", "https://notificaties-api.vng.cloud/api/v1/")
class NotificationsConfig(ClientConfig):
    class Meta:
        verbose_name = _("Notificatiescomponentconfiguratie")

    def get_auth(self) -> ClientAuth:
        auth = APICredential.get_auth(
            self.api_root, scopes=[SCOPE_NOTIFICATIES_PUBLICEREN_LABEL]
        )
        return auth


class Subscription(models.Model):
    """
    A single subscription.

    TODO: on change/update, update the subscription
    """

    config = models.ForeignKey("NotificationsConfig", on_delete=models.CASCADE)

    callback_url = models.URLField(
        _("callback url"), help_text=_("Where to send the notifications (webhook url)")
    )
    client_id = models.CharField(
        _("client ID"),
        max_length=50,
        help_text=_("Client ID to construct the auth token"),
    )
    secret = models.CharField(
        _("client secret"),
        max_length=50,
        help_text=_("Secret to construct the auth token"),
    )
    channels = ArrayField(
        models.CharField(max_length=100),
        verbose_name=_("channels"),
        help_text=_("Comma-separated list of channels to subscribe to"),
    )

    _subscription = models.URLField(
        _("NC subscription"),
        blank=True,
        editable=False,
        help_text=_("Subscription as it is known in the NC"),
    )

    class Meta:
        verbose_name = _("Webhook subscription")
        verbose_name_plural = _("Webhook subscriptions")

    def __str__(self):
        return f"{', '.join(self.channels)} - {self.callback_url}"

    def register(self) -> None:
        """
        Registers the webhook with the notification component.
        """
        dummy_detail_url = urljoin(self.config.api_root, f"foo/{uuid.uuid4()}")
        client = Client.from_url(dummy_detail_url)

        # This authentication is to create a subscription at the NC.
        client.auth = APICredential.get_auth(
            self.config.api_root, scopes=[SCOPE_NOTIFICATIES_CONSUMEREN_LABEL]
        )

        # This authentication is for the NC to call us. Thus, it's *not* for
        # calling the NC to create a subscription.
        self_auth = ClientAuth(
            client_id=self.client_id,
            secret=self.secret,
            scopes=[SCOPE_NOTIFICATIES_PUBLICEREN_LABEL],
        )
        data = {
            "callbackUrl": self.callback_url,
            "auth": self_auth.credentials()["Authorization"],
            "kanalen": [
                {
                    "naam": channel,
                    # FIXME: You need to be able to configure these.
                    "filters": {},
                }
                for channel in self.channels
            ],
        }

        # register the subscriber
        subscriber = client.create("abonnement", data=data)

        self._subscription = subscriber["url"]
        self.save(update_fields=["_subscription"])



# Generated by Django 2.2.2 on 2019-07-12 14:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("authorizations", "0009_update_enums")]

    operations = [
        migrations.AlterField(
            model_name="authorizationsconfig",
            name="api_root",
            field=models.URLField(
                default="https://autorisaties-api.vng.cloud/api/v1",
                unique=True,
                verbose_name="api root",
            ),
        )
    ]



# Generated by Django 2.2.2 on 2019-06-06 12:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("audittrails", "0007_auto_20190522_0916")]

    operations = [
        migrations.AddField(
            model_name="audittrail",
            name="resource_weergave",
            field=models.CharField(
                default="",
                help_text="Vriendelijke identificatie van het object",
                max_length=200,
            ),
            preserve_default=False,
        )
    ]



from rest_framework import serializers

from ...constants import CommonResourceAction, ComponentTypes
from ...serializers import GegevensGroepSerializer, add_choice_values_help_text
from ..models import AuditTrail


class WijzigingenSerializer(GegevensGroepSerializer):
    class Meta:
        model = AuditTrail
        gegevensgroep = "wijzigingen"


class AuditTrailSerializer(serializers.ModelSerializer):
    wijzigingen = WijzigingenSerializer()

    class Meta:
        model = AuditTrail
        fields = (
            "uuid",
            "bron",
            "applicatie_id",
            "applicatie_weergave",
            "gebruikers_id",
            "gebruikers_weergave",
            "actie",
            "actie_weergave",
            "resultaat",
            "hoofd_object",
            "resource",
            "resource_url",
            "toelichting",
            "resource_weergave",
            "aanmaakdatum",
            "wijzigingen",
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        value_display_mapping = add_choice_values_help_text(ComponentTypes)
        self.fields["bron"].help_text += f"\n\n{value_display_mapping}"

        # Indicate that the values for AuditTrail.actie are not limited to
        # the CommonResourceActions
        custom_msg = """De bekende waardes voor dit veld zijn hieronder aangegeven, \
                        maar andere waardes zijn ook toegestaan"""

        value_display_mapping = add_choice_values_help_text(CommonResourceAction)
        self.fields["actie"].help_text += f"\n\n{custom_msg}\n\n{value_display_mapping}"



# Generated by Django 2.2.1 on 2019-06-20 08:49

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("testapp", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="Group",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                )
            ],
        ),
        migrations.AddField(
            model_name="person",
            name="group",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="testapp.Group",
            ),
        ),
    ]



from .kanalen import Kanaal


def notification_documentation(kanaal: Kanaal):
    """
    Generate notification documentation for an OpenAPI specification containing
    the relevant resources and actions for a given KANAAL
    """
    doc = f"""Deze API publiceert notificaties op het kanaal `{kanaal.label}`.

{kanaal.description}

**Resources en acties**
"""
    for resource, actions in kanaal.get_usage():
        doc += f"""- `{resource}`: {', '.join(actions)}\n"""
    return doc



class Audit:
    def __init__(self, component_name: str, main_resource: str):
        self.component_name = component_name
        self.main_resource = main_resource

    def __repr__(self):
        cls_name = self.__class__.__name__
        return f"{cls_name}(component_name={self.component_name}, main_resource={self.main_resource})"



import logging

from django.conf import settings
from django.contrib.sites.models import Site
from django.core.management.base import BaseCommand
from django.urls import reverse
from django.utils.module_loading import import_string

from ....models import APICredential
from ...constants import SCOPE_NOTIFICATIES_PUBLICEREN_LABEL
from ...kanalen import KANAAL_REGISTRY
from ...models import NotificationsConfig

logger = logging.getLogger(__name__)


class KanaalExists(Exception):
    pass


def create_kanaal(api_root: str, kanaal: str) -> None:
    """
    Create a kanaal, if it doesn't exist yet.
    """
    Client = import_string(settings.ZDS_CLIENT_CLASS)

    if not api_root.endswith("/"):
        api_root = f"{api_root}/"

    client = Client.from_url(api_root)
    client.base_url = api_root
    client.auth = APICredential.get_auth(
        api_root, scopes=[SCOPE_NOTIFICATIES_PUBLICEREN_LABEL]
    )

    # look up the exchange in the registry
    _kanaal = next(k for k in KANAAL_REGISTRY if k.label == kanaal)

    kanalen = client.list("kanaal", query_params={"naam": kanaal})
    if kanalen:
        raise KanaalExists()

    # build up own documentation URL
    domain = Site.objects.get_current().domain
    protocol = "https" if settings.IS_HTTPS else "http"
    documentation_url = (
        f"{protocol}://{domain}{reverse('notifications:kanalen')}#{kanaal}"
    )

    client.create(
        "kanaal",
        {
            "naam": kanaal,
            "documentatieLink": documentation_url,
            "filters": list(_kanaal.kenmerken),
        },
    )


class Command(BaseCommand):
    help = "Create kanaal in notification component"

    def add_arguments(self, parser):
        parser.add_argument("kanaal", nargs="?", type=str, help="Name of kanaal")
        parser.add_argument(
            "--nc-api-root",
            help="API root of the NC, default value taken from notifications config",
        )

    def handle(self, **options):
        config = NotificationsConfig.get_solo()

        # use CLI arg or fall back to database config
        api_root = options["nc_api_root"] or config.api_root

        # use CLI arg or fall back to setting
        kanaal = options["kanaal"] or settings.NOTIFICATIONS_KANAAL

        try:
            create_kanaal(api_root, kanaal)
            self.stdout.write(f"Registered kanaal '{kanaal}' with {api_root}")
        except KanaalExists:
            self.stderr.write(f"Kanaal '{kanaal}' already exists within {api_root}")



from vng_api_common.scopes import Scope

SCOPE_AUDITTRAILS_LEZEN = Scope(
    "audittrails.lezen",
    description="""
**Laat toe om**:

* audittrails op te lijsten
* audittrail details op te vragen
""",
)



# Generated by Django 2.0.13 on 2019-05-03 09:41

from django.db import migrations, models

import vng_api_common.fields


class Migration(migrations.Migration):

    dependencies = [("authorizations", "0003_auto_20190502_0409")]

    operations = [
        migrations.AlterField(
            model_name="autorisatie",
            name="max_vertrouwelijkheidaanduiding",
            field=vng_api_common.fields.VertrouwelijkheidsAanduidingField(
                blank=True,
                choices=[
                    ("openbaar", "OPENBAAR"),
                    ("beperkt openbaar", "BEPERKT OPENBAAR"),
                    ("intern", "INTERN"),
                    ("zaakvertrouwelijk", "ZAAKVERTROUWELIJK"),
                    ("vertrouwelijk", "VERTROUWELIJK"),
                    ("confidentieel", "CONFIDENTIEEL"),
                    ("geheim", "GEHEIM"),
                    ("zeer geheim", "ZEER GEHEIM"),
                ],
                help_text="Maximum level of confidentiality that is allowed",
                max_length=20,
            ),
        ),
        migrations.AlterField(
            model_name="autorisatie",
            name="zaaktype",
            field=models.URLField(
                blank=True,
                help_text="Url of the zaaktype that is allowed",
                max_length=1000,
            ),
        ),
    ]



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from solo.admin import SingletonModelAdmin

from .models import NotificationsConfig, Subscription


class SubscriptionInline(admin.TabularInline):
    model = Subscription
    extra = 0


@admin.register(NotificationsConfig)
class NotificationsConfigAdmin(SingletonModelAdmin):
    list_display = ("api_root", "subscriptions")
    inlines = [SubscriptionInline]

    def subscriptions(self, obj):
        urls = obj.exclude(subscription_set___subscription="").values_list(
            "subscription_set___subscription"
        )
        return ", ".join(urls)


def register_webhook(modeladmin, request, queryset):
    for sub in queryset:
        if sub._subscription:
            continue
        sub.register()


register_webhook.short_description = _("Register the webhooks")  # noqa


@admin.register(Subscription)
class SubscriptionAdmin(admin.ModelAdmin):
    list_display = ("callback_url", "channels", "_subscription")
    actions = [register_webhook]



"""
Introspect polymorphic resources

Bulk of the code taken from https://github.com/axnsan12/drf-yasg/issues/100
"""
from drf_yasg import openapi
from drf_yasg.errors import SwaggerGenerationError
from drf_yasg.inspectors.base import NotHandled
from drf_yasg.inspectors.field import (
    CamelCaseJSONFilter,
    ReferencingSerializerInspector,
)

from ..polymorphism import PolymorphicSerializer
from ..utils import underscore_to_camel


class PolymorphicSerializerInspector(
    CamelCaseJSONFilter, ReferencingSerializerInspector
):
    def field_to_swagger_object(
        self, field, swagger_object_type, use_references, **kwargs
    ):
        SwaggerType, ChildSwaggerType = self._get_partial_types(
            field, swagger_object_type, use_references, **kwargs
        )

        if not isinstance(field, PolymorphicSerializer):
            return NotHandled

        if not getattr(field, "discriminator", None):
            raise SwaggerGenerationError(
                "'PolymorphicSerializer' derived serializers need to have 'discriminator' set"
            )

        base_schema_ref = super().field_to_swagger_object(
            field, swagger_object_type, use_references, **kwargs
        )
        if not isinstance(base_schema_ref, openapi.SchemaRef):
            raise SwaggerGenerationError(
                "discriminator inheritance requires model references"
            )

        base_schema = base_schema_ref.resolve(self.components)  # type: openapi.Schema
        base_schema.discriminator = underscore_to_camel(
            field.discriminator.discriminator_field
        )

        for value, serializer in field.discriminator.mapping.items():
            if serializer is None:
                allof_derived = openapi.Schema(
                    type=openapi.TYPE_OBJECT, all_of=[base_schema_ref]
                )
            else:
                derived_ref = self.probe_field_inspectors(
                    serializer, openapi.Schema, use_references=True
                )
                if not isinstance(derived_ref, openapi.SchemaRef):
                    raise SwaggerGenerationError(
                        "discriminator inheritance requies model references"
                    )

                allof_derived = openapi.Schema(
                    type=openapi.TYPE_OBJECT, all_of=[base_schema_ref, derived_ref]
                )
            if not self.components.has(value, scope=openapi.SCHEMA_DEFINITIONS):
                self.components.set(
                    value, allof_derived, scope=openapi.SCHEMA_DEFINITIONS
                )

        return base_schema_ref



from rest_framework.generics import CreateAPIView
from rest_framework.request import Request
from rest_framework.views import APIView

from ..constants import ComponentTypes
from ..permissions import BaseAuthRequired, bypass_permissions, get_required_scopes


class JWTCreatePermission(BaseAuthRequired):
    def has_permission(self, request: Request, view: APIView) -> bool:
        if bypass_permissions(request):
            return True

        if not isinstance(view, CreateAPIView):
            return False

        if not hasattr(view, "action"):
            view.action = "create"

        scopes_required = get_required_scopes(view)
        return request.jwt_auth.has_auth(scopes_required, component=ComponentTypes.ac)



# Generated by Django 2.1.8 on 2019-05-06 08:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("authorizations", "0004_auto_20190503_0941")]

    operations = [
        migrations.AlterField(
            model_name="authorizationsconfig",
            name="component",
            field=models.CharField(
                choices=[
                    ("AC", "Autorisatiecomponent"),
                    ("NRC", "Notificatierouteringcomponent"),
                    ("ZRC", "Zaakregistratiecomponent"),
                    ("ZTC", "Zaaktypecatalogus"),
                    ("DRC", "Documentregistratiecomponent"),
                    ("BRC", "Besluitregistratiecomponent"),
                ],
                default="ZRC",
                max_length=50,
                verbose_name="component",
            ),
        ),
        migrations.AlterField(
            model_name="autorisatie",
            name="component",
            field=models.CharField(
                choices=[
                    ("AC", "Autorisatiecomponent"),
                    ("NRC", "Notificatierouteringcomponent"),
                    ("ZRC", "Zaakregistratiecomponent"),
                    ("ZTC", "Zaaktypecatalogus"),
                    ("DRC", "Documentregistratiecomponent"),
                    ("BRC", "Besluitregistratiecomponent"),
                ],
                help_text="Name of the component to authorize",
                max_length=50,
            ),
        ),
    ]



from django.utils.translation import ugettext_lazy as _

from djchoices import ChoiceItem, DjangoChoices

BSN_LENGTH = 9
RSIN_LENGTH = 9

VERSION_HEADER = "API-version"

SCOPE_NOTIFICATIES_PUBLICEREN_LABEL = "notificaties.scopes.publiceren"

# see https://docs.nlx.io/further-reading/transaction-logs/
HEADER_APPLICATION = "X-NLX-Request-Application-Id"
HEADER_USER_ID = "X-NLX-Request-User-Id"
HEADER_AUDIT = "X-Audit-Toelichting"

FILTER_URL_DID_NOT_RESOLVE = "NO_MATCHING_OBJECT"


class VertrouwelijkheidsAanduiding(DjangoChoices):
    openbaar = ChoiceItem("openbaar", "Openbaar")
    beperkt_openbaar = ChoiceItem("beperkt_openbaar", "Beperkt openbaar")
    intern = ChoiceItem("intern", "Intern")
    zaakvertrouwelijk = ChoiceItem("zaakvertrouwelijk", "Zaakvertrouwelijk")
    vertrouwelijk = ChoiceItem("vertrouwelijk", "Vertrouwelijk")
    confidentieel = ChoiceItem("confidentieel", "Confidentieel")
    geheim = ChoiceItem("geheim", "Geheim")
    zeer_geheim = ChoiceItem("zeer_geheim", "Zeer geheim")


class RolOmschrijving(DjangoChoices):
    adviseur = ChoiceItem(
        "adviseur",
        "Adviseur",
        description="Kennis in dienst stellen van de behandeling van (een deel van) een zaak.",
    )
    behandelaar = ChoiceItem(
        "behandelaar",
        "Behandelaar",
        description="De vakinhoudelijke behandeling doen van (een deel van) een zaak.",
    )
    belanghebbende = ChoiceItem(
        "belanghebbende",
        "Belanghebbende",
        description="Vanuit eigen en objectief belang rechtstreeks betrokken "
        "zijn bij de behandeling en/of de uitkomst van een zaak.",
    )
    beslisser = ChoiceItem(
        "beslisser",
        "Beslisser",
        description="Nemen van besluiten die voor de uitkomst van een zaak noodzakelijk zijn.",
    )
    initiator = ChoiceItem(
        "initiator",
        "Initiator",
        description="Aanleiding geven tot de start van een zaak ..",
    )
    klantcontacter = ChoiceItem(
        "klantcontacter",
        "Klantcontacter",
        description="Het eerste aanspreekpunt zijn voor vragen van burgers en bedrijven ..",
    )
    zaakcoordinator = ChoiceItem(
        "zaakcoordinator",
        "Zaakcoördinator",
        description="Er voor zorg dragen dat de behandeling van de zaak in samenhang "
        "uitgevoerd wordt conform de daarover gemaakte afspraken.",
    )
    medeinitiator = ChoiceItem("mede_initiator", "Mede-initiator", description="")


class RolTypes(DjangoChoices):
    natuurlijk_persoon = ChoiceItem("natuurlijk_persoon", "Natuurlijk persoon")
    niet_natuurlijk_persoon = ChoiceItem(
        "niet_natuurlijk_persoon", "Niet-natuurlijk persoon"
    )
    vestiging = ChoiceItem("vestiging", "Vestiging")
    organisatorische_eenheid = ChoiceItem(
        "organisatorische_eenheid", "Organisatorische eenheid"
    )
    medewerker = ChoiceItem("medewerker", "Medewerker")


class ObjectTypes(DjangoChoices):
    besluit = ChoiceItem("besluit", _("Besluit"))
    zaak = ChoiceItem("zaak", _("Zaak"))


class Archiefnominatie(DjangoChoices):
    blijvend_bewaren = ChoiceItem(
        "blijvend_bewaren",
        _(
            "Het zaakdossier moet bewaard blijven en op de Archiefactiedatum overgedragen worden naar een "
            "archiefbewaarplaats."
        ),
    )
    vernietigen = ChoiceItem(
        "vernietigen",
        _("Het zaakdossier moet op of na de Archiefactiedatum vernietigd worden."),
    )


class Archiefstatus(DjangoChoices):
    nog_te_archiveren = ChoiceItem(
        "nog_te_archiveren",
        _("De zaak cq. het zaakdossier is nog niet als geheel gearchiveerd."),
    )
    gearchiveerd = ChoiceItem(
        "gearchiveerd",
        _(
            "De zaak cq. het zaakdossier is als geheel niet-wijzigbaar bewaarbaar gemaakt."
        ),
    )
    gearchiveerd_procestermijn_onbekend = ChoiceItem(
        "gearchiveerd_procestermijn_onbekend",
        _(
            "De zaak cq. het zaakdossier is als geheel niet-wijzigbaar bewaarbaar gemaakt maar de vernietigingsdatum "
            "kan nog niet bepaald worden."
        ),
    )
    # After deliberation this element was removed because "vernietigd" means
    # it's really gone and the status wouldn't make sense:
    #
    # vernietigd = ChoiceItem('vernietigd',
    #     _("De zaak cq. het zaakdossier is vernietigd.")
    # )
    overgedragen = ChoiceItem(
        "overgedragen",
        _("De zaak cq. het zaakdossier is overgebracht naar een archiefbewaarplaats."),
    )


class BrondatumArchiefprocedureAfleidingswijze(DjangoChoices):
    afgehandeld = ChoiceItem(
        "afgehandeld",
        _("Afgehandeld"),
        description=_(
            "De termijn start op de datum waarop de zaak is "
            "afgehandeld (ZAAK.Einddatum in het RGBZ)."
        ),
    )
    ander_datumkenmerk = ChoiceItem(
        "ander_datumkenmerk",
        _("Ander datumkenmerk"),
        description=_(
            "De termijn start op de datum die is vastgelegd in een "
            "ander datumveld dan de datumvelden waarop de overige "
            "waarden (van deze attribuutsoort) betrekking hebben. "
            "`Objecttype`, `Registratie` en `Datumkenmerk` zijn niet "
            "leeg."
        ),
    )
    eigenschap = ChoiceItem(
        "eigenschap",
        _("Eigenschap"),
        description=_(
            "De termijn start op de datum die vermeld is in een "
            "zaaktype-specifieke eigenschap (zijnde een `datumveld`). "
            "`ResultaatType.ZaakType` heeft een `Eigenschap`; "
            "`Objecttype`, en `Datumkenmerk` zijn niet leeg."
        ),
    )
    gerelateerde_zaak = ChoiceItem(
        "gerelateerde_zaak",
        _("Gerelateerde zaak"),
        description=_(
            "De termijn start op de datum waarop de gerelateerde "
            "zaak is afgehandeld (`ZAAK.Einddatum` of "
            "`ZAAK.Gerelateerde_zaak.Einddatum` in het RGBZ). "
            "`ResultaatType.ZaakType` heeft gerelateerd `ZaakType`"
        ),
    )
    hoofdzaak = ChoiceItem(
        "hoofdzaak",
        _("Hoofdzaak"),
        description=_(
            "De termijn start op de datum waarop de gerelateerde "
            "zaak is afgehandeld, waarvan de zaak een deelzaak is "
            "(`ZAAK.Einddatum` van de hoofdzaak in het RGBZ). "
            "ResultaatType.ZaakType is deelzaaktype van ZaakType."
        ),
    )
    ingangsdatum_besluit = ChoiceItem(
        "ingangsdatum_besluit",
        _("Ingangsdatum besluit"),
        description=_(
            "De termijn start op de datum waarop het besluit van "
            "kracht wordt (`BESLUIT.Ingangsdatum` in het RGBZ).	"
            "ResultaatType.ZaakType heeft relevant BesluitType"
        ),
    )
    termijn = ChoiceItem(
        "termijn",
        _("Termijn"),
        description=_(
            "De termijn start een vast aantal jaren na de datum "
            "waarop de zaak is afgehandeld (`ZAAK.Einddatum` in het "
            "RGBZ)."
        ),
    )
    vervaldatum_besluit = ChoiceItem(
        "vervaldatum_besluit",
        _("Vervaldatum besluit"),
        description=_(
            "De termijn start op de dag na de datum waarop het "
            "besluit vervalt (`BESLUIT.Vervaldatum` in het RGBZ). "
            "ResultaatType.ZaakType heeft relevant BesluitType"
        ),
    )
    zaakobject = ChoiceItem(
        "zaakobject",
        _("Zaakobject"),
        description=_(
            "De termijn start op de einddatum geldigheid van het "
            "zaakobject waarop de zaak betrekking heeft "
            "(bijvoorbeeld de overlijdendatum van een Persoon). "
            "M.b.v. de attribuutsoort `Objecttype` wordt vastgelegd "
            "om welke zaakobjecttype het gaat; m.b.v. de "
            "attribuutsoort `Datumkenmerk` wordt vastgelegd welke "
            "datum-attribuutsoort van het zaakobjecttype het betreft."
        ),
    )


class ZaakobjectTypes(DjangoChoices):
    adres = ChoiceItem("adres", "Adres")
    besluit = ChoiceItem("besluit", "Besluit")
    buurt = ChoiceItem("buurt", "Buurt")
    enkelvoudig_document = ChoiceItem("enkelvoudig_document", "Enkelvoudig document")
    gemeente = ChoiceItem("gemeente", "Gemeente")
    gemeentelijke_openbare_ruimte = ChoiceItem(
        "gemeentelijke_openbare_ruimte", "Gemeentelijke openbare ruimte"
    )
    huishouden = ChoiceItem("huishouden", "Huishouden")
    inrichtingselement = ChoiceItem("inrichtingselement", "Inrichtingselement")
    kadastrale_onroerende_zaak = ChoiceItem(
        "kadastrale_onroerende_zaak", "Kadastrale onroerende zaak"
    )
    kunstwerkdeel = ChoiceItem("kunstwerkdeel", "Kunstwerkdeel")
    maatschappelijke_activiteit = ChoiceItem(
        "maatschappelijke_activiteit", "Maatschappelijke activiteit"
    )
    medewerker = ChoiceItem("medewerker", "Medewerker")
    natuurlijk_persoon = ChoiceItem("natuurlijk_persoon", "Natuurlijk persoon")
    niet_natuurlijk_persoon = ChoiceItem(
        "niet_natuurlijk_persoon", "Niet-natuurlijk persoon"
    )
    openbare_ruimte = ChoiceItem("openbare_ruimte", "Openbare ruimte")
    organisatorische_eenheid = ChoiceItem(
        "organisatorische_eenheid", "Organisatorische eenheid"
    )
    pand = ChoiceItem("pand", "Pand")
    spoorbaandeel = ChoiceItem("spoorbaandeel", "Spoorbaandeel")
    status = ChoiceItem("status", "Status")
    terreindeel = ChoiceItem("terreindeel", "Terreindeel")
    terrein_gebouwd_object = ChoiceItem(
        "terrein_gebouwd_object", "Terrein gebouwd object"
    )
    vestiging = ChoiceItem("vestiging", "Vestiging")
    waterdeel = ChoiceItem("waterdeel", "Waterdeel")
    wegdeel = ChoiceItem("wegdeel", "Wegdeel")
    wijk = ChoiceItem("wijk", "Wijk")
    woonplaats = ChoiceItem("woonplaats", "Woonplaats")
    woz_deelobject = ChoiceItem("woz_deelobject", "Woz deel object")
    woz_object = ChoiceItem("woz_object", "Woz object")
    woz_waarde = ChoiceItem("woz_waarde", "Woz waarde")
    zakelijk_recht = ChoiceItem("zakelijk_recht", "Zakelijk recht")
    overige = ChoiceItem("overige", "Overige")


class ComponentTypes(DjangoChoices):
    ac = ChoiceItem("ac", "Autorisatiecomponent")
    nrc = ChoiceItem("nrc", "Notificatierouteringcomponent")
    zrc = ChoiceItem("zrc", "Zaakregistratiecomponent")
    ztc = ChoiceItem("ztc", "Zaaktypecatalogus")
    drc = ChoiceItem("drc", "Documentregistratiecomponent")
    brc = ChoiceItem("brc", "Besluitregistratiecomponent")


class CommonResourceAction(DjangoChoices):
    create = ChoiceItem("create", _("Object aangemaakt"))
    list = ChoiceItem("list", _("Lijst van objecten opgehaald"))
    retrieve = ChoiceItem("retrieve", _("Object opgehaald"))
    destroy = ChoiceItem("destroy", _("Object verwijderd"))
    update = ChoiceItem("update", _("Object bijgewerkt"))
    partial_update = ChoiceItem("partial_update", _("Object deels bijgewerkt"))


class RelatieAarden(DjangoChoices):
    hoort_bij = ChoiceItem("hoort_bij", _("Hoort bij, omgekeerd: kent"))
    legt_vast = ChoiceItem(
        "legt_vast", _("Legt vast, omgekeerd: kan vastgelegd zijn als")
    )

    @classmethod
    def from_object_type(cls, object_type: str) -> str:
        if object_type == ObjectTypes.zaak:
            return cls.hoort_bij

        if object_type == ObjectTypes.besluit:
            return cls.legt_vast

        raise ValueError(f"Unknown object_type '{object_type}'")



"""
Documentation strings to be included in the API schema.
"""

DOC_AUTH_JWT = """
**Autorisatie**

Deze API vereist autorisatie. Je kan de
[token-tool](https://zaken-auth.vng.cloud/) gebruiken om een token te
genereren.

_Zelf een token genereren_

De tokens die gebruikt worden voor autorisatie zijn [jwt.io][JWT's] (JSON web
token). In de API calls moeten deze gebruikt worden in de `Authorization`
header:

```
Authorization: Bearer <token>
```

Om een JWT te genereren heb je een `client ID` en een `secret` nodig. Het JWT
moet gebouwd worden volgens het `HS256` algoritme. De vereiste payload is:

```json
{
    "iss": "<client ID>",
    "iat": 1572863906,
    "client_id": "<client ID>",
    "user_id": "<user identifier>",
    "user_representation": "<user representation>"
}
```

Als `issuer` gebruik je dus je eigen client ID. De `iat` timestamp is een
UNIX-timestamp die aangeeft op welk moment het token gegenereerd is.

`user_id` en `user_representation` zijn nodig voor de audit trails. Het zijn
vrije velden met als enige beperking dat de lengte maximaal de lengte van
de overeenkomstige velden in de audit trail resources is (zie rest API spec).
"""



# Generated by Django 2.1.1 on 2019-07-12 15:53

from django.db import migrations
from django.db.models import F, Value
from django.db.models.functions import Lower, Replace


def forward(apps, schema_editor):
    AuthorizationsConfig = apps.get_model("authorizations", "AuthorizationsConfig")
    AuthorizationsConfig.objects.all().update(component=Lower(F("component")))

    Autorisatie = apps.get_model("authorizations", "Autorisatie")
    Autorisatie.objects.all().update(
        component=Lower(F("component")),
        max_vertrouwelijkheidaanduiding=Lower(
            Replace(F("max_vertrouwelijkheidaanduiding"), Value(" "), Value("_"))
        ),
    )


class Migration(migrations.Migration):

    dependencies = [("authorizations", "0008_auto_20190712_1541")]

    operations = [migrations.RunPython(forward, migrations.RunPython.noop)]



# Generated by Django 2.1.8 on 2019-05-02 04:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("authorizations", "0002_authorizationsconfig")]

    operations = [
        migrations.AlterField(
            model_name="authorizationsconfig",
            name="api_root",
            field=models.URLField(
                default="https://ref.tst.vng.cloud/ac/api/v1",
                unique=True,
                verbose_name="api root",
            ),
        )
    ]



import time
from typing import List, Optional

import jwt
from rest_framework import status

from ..authorizations.models import Applicatie, AuthorizationsConfig, Autorisatie
from ..constants import VertrouwelijkheidsAanduiding
from ..models import JWTSecret


def generate_jwt(scopes: list, secret: str = "letmein", zaaktypes: list = None) -> str:
    scope_labels = sum((_get_scope_labels(scope) for scope in scopes), [])
    payload = {
        # standard claims
        "iss": "testsuite",
        "iat": int(time.time()),
        # custom claims
        "zds": {"scopes": scope_labels, "zaaktypes": zaaktypes or []},
    }
    headers = {"client_identifier": "testsuite"}
    encoded = jwt.encode(payload, secret, headers=headers, algorithm="HS256")
    encoded = encoded.decode("ascii")
    return f"Bearer {encoded}"


def _get_scope_labels(scope) -> list:
    if not scope.children:
        return [scope.label]

    labels = []
    for child in scope.children:
        labels += _get_scope_labels(child)
    return sorted(set(labels))


class AuthCheckMixin:
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        JWTSecret.objects.get_or_create(
            identifier="testsuite", defaults={"secret": "letmein"}
        )

    def assertForbidden(self, url, method="get", request_kwargs=None):
        """
        Assert that an appropriate scope is required.
        """
        do_request = getattr(self.client, method)
        request_kwargs = request_kwargs or {}

        with self.subTest(case="JWT missing"):
            response = do_request(url, **request_kwargs)

            self.assertEqual(
                response.status_code, status.HTTP_403_FORBIDDEN, response.data
            )

        with self.subTest(case="Invalid JWT structure"):
            invalid_jwt = generate_jwt_auth("testsuite", "letmein")[:-10]
            self.client.credentials(HTTP_AUTHORIZATION=invalid_jwt)

            response = do_request(url, **request_kwargs)

            self.assertEqual(
                response.status_code, status.HTTP_403_FORBIDDEN, response.data
            )

    def assertForbiddenWithCorrectScope(
        self, url: str, scopes: list, method="get", request_kwargs=None, **extra_claims
    ):

        do_request = getattr(self.client, method)
        request_kwargs = request_kwargs or {}

        jwt = generate_jwt(scopes=scopes, **extra_claims)
        self.client.credentials(HTTP_AUTHORIZATION=jwt)

        response = do_request(url, **request_kwargs)

        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)


class JWTScopesMixin:

    scopes = None
    zaaktypes = None

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        JWTSecret.objects.get_or_create(
            identifier="testsuite", defaults={"secret": "letmein"}
        )

    def setUp(self):
        super().setUp()

        if self.scopes is not None:
            token = generate_jwt(
                scopes=self.scopes, zaaktypes=self.zaaktypes or [], secret="letmein"
            )
            self.client.credentials(HTTP_AUTHORIZATION=token)


# tools fot testing with new authorization format
def generate_jwt_auth(
    client_id, secret, user_id="test_user_id", user_representation="Test User"
):
    """
    Generate a JWT suitable for the second version of the AC-based auth.
    """
    payload = {
        # standard claims
        "iss": "testsuite",
        "iat": int(time.time()),
        # custom
        "client_id": client_id,
        "user_id": user_id,
        "user_representation": user_representation,
    }
    encoded = jwt.encode(payload, secret, algorithm="HS256")
    encoded = encoded.decode("ascii")
    return f"Bearer {encoded}"


class JWTAuthMixin:
    """
    Configure the local auth cache.

    Creates the local auth objects for permission checks, as if you're talking
    to a real AC behind the scenes.
    """

    client_id = "testsuite"
    secret = "letmein"

    user_id = "test_user_id"
    user_representation = "Test User"

    scopes = None
    heeft_alle_autorisaties = False
    zaaktype = None
    informatieobjecttype = None
    besluittype = None
    max_vertrouwelijkheidaanduiding = VertrouwelijkheidsAanduiding.zeer_geheim

    @staticmethod
    def _create_credentials(
        client_id: str,
        secret: str,
        heeft_alle_autorisaties: bool,
        max_vertrouwelijkheidaanduiding: str,
        scopes: Optional[List[str]] = None,
        zaaktype: Optional[str] = None,
        informatieobjecttype: Optional[str] = None,
        besluittype: Optional[str] = None,
    ):
        JWTSecret.objects.get_or_create(
            identifier=client_id, defaults={"secret": secret}
        )

        config = AuthorizationsConfig.get_solo()

        applicatie = Applicatie.objects.create(
            client_ids=[client_id],
            label="for test",
            heeft_alle_autorisaties=heeft_alle_autorisaties,
        )

        if heeft_alle_autorisaties is False:
            autorisatie = Autorisatie.objects.create(
                applicatie=applicatie,
                component=config.component,
                scopes=scopes or [],
                zaaktype=zaaktype or "",
                informatieobjecttype=informatieobjecttype or "",
                besluittype=besluittype or "",
                max_vertrouwelijkheidaanduiding=max_vertrouwelijkheidaanduiding,
            )
        else:
            autorisatie = None

        return applicatie, autorisatie

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        applicatie, autorisatie = cls._create_credentials(
            cls.client_id,
            cls.secret,
            heeft_alle_autorisaties=cls.heeft_alle_autorisaties,
            scopes=cls.scopes,
            zaaktype=cls.zaaktype,
            informatieobjecttype=cls.informatieobjecttype,
            besluittype=cls.besluittype,
            max_vertrouwelijkheidaanduiding=cls.max_vertrouwelijkheidaanduiding,
        )
        cls.applicatie = applicatie
        cls.autorisatie = autorisatie

    def setUp(self):
        super().setUp()

        token = generate_jwt_auth(
            client_id=self.client_id,
            secret=self.secret,
            user_id=self.user_id,
            user_representation=self.user_representation,
        )
        self.client.credentials(HTTP_AUTHORIZATION=token)



# Generated by Django 2.0.13 on 2019-05-17 08:44

import uuid

import django.contrib.postgres.fields.jsonb
import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("audittrails", "0002_auto_20190516_0830")]

    operations = [
        migrations.AlterField(
            model_name="audittrail",
            name="aanmaakdatum",
            field=models.DateTimeField(
                auto_now=True, help_text="De datum waarop de handeling is gedaan"
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="actie",
            field=models.CharField(help_text="De uitgevoerde handeling", max_length=50),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="actie_weergave",
            field=models.CharField(
                blank=True, help_text="Vriendelijke naam van de actie", max_length=200
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="bron",
            field=models.CharField(
                help_text="De naam van het component waar de wijziging in is gedaan",
                max_length=50,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="hoofd_object",
            field=models.URLField(
                help_text="De URL naar het hoofdobject van een component",
                max_length=1000,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="nieuw",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                help_text="Volledige JSON body van het object na de actie",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="oud",
            field=django.contrib.postgres.fields.jsonb.JSONField(
                encoder=django.core.serializers.json.DjangoJSONEncoder,
                help_text="Volledige JSON body van het object zoals dat bestond voordat de actie heeft plaatsgevonden",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="resource",
            field=models.CharField(
                help_text="Het type resource waarop de actie gebeurde", max_length=50
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="resource_url",
            field=models.URLField(help_text="De URL naar het object", max_length=1000),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="resultaat",
            field=models.IntegerField(
                help_text="HTTP status code van de API response van de uitgevoerde handeling"
            ),
        ),
        migrations.AlterField(
            model_name="audittrail",
            name="uuid",
            field=models.UUIDField(
                default=uuid.uuid4,
                help_text="Unieke identificatie van de audit regel",
                unique=True,
            ),
        ),
    ]



from django.contrib import admin

from solo.admin import SingletonModelAdmin

from .models import Applicatie, AuthorizationsConfig, Autorisatie


@admin.register(AuthorizationsConfig)
class AuthorizationsConfigAdmin(SingletonModelAdmin):
    list_display = ("api_root", "component")


@admin.register(Autorisatie)
class AutorisatieAdmin(admin.ModelAdmin):
    list_display = (
        "applicatie",
        "component",
        "zaaktype",
        "scopes",
        "max_vertrouwelijkheidaanduiding",
    )


class AutorisatieInline(admin.TabularInline):
    model = Autorisatie


@admin.register(Applicatie)
class ApplicatieAdmin(admin.ModelAdmin):
    list_display = ("uuid", "client_ids", "label", "heeft_alle_autorisaties")
    readonly_fields = ("uuid",)
    inlines = (AutorisatieInline,)



from rest_framework import serializers
from testapp.models import Group, Person

from vng_api_common.serializers import GegevensGroepSerializer


class AddressSerializer(GegevensGroepSerializer):
    class Meta:
        model = Person
        gegevensgroep = "address"


class PersonSerializer(serializers.ModelSerializer):
    address = AddressSerializer(allow_null=True)

    class Meta:
        model = Person
        fields = ("address", "name")


class PersonSerializer2(serializers.ModelSerializer):
    address = AddressSerializer(allow_null=True, required=False)

    class Meta:
        model = Person
        fields = ("address", "name")


class GroupSerializer(serializers.ModelSerializer):
    person = PersonSerializer(many=True)

    class Meta:
        model = Group
        fields = ("person",)



from typing import List

from django.utils.translation import ugettext_lazy as _

from rest_framework.exceptions import ErrorDetail
from rest_framework.serializers import ValidationError

from ..constants import ComponentTypes
from ..utils import underscore_to_camel
from .models import Applicatie


class UniqueClientIDValidator:
    code = "clientId-exists"
    message = _(
        "The clientID(s) {client_id} are already used in application(s) {app_id}"
    )

    def set_context(self, serializer_field):
        """
        This hook is called by the serializer instance,
        prior to the validation call being made.
        """
        # Determine the existing instance, if this is an update operation.
        self.instance = getattr(serializer_field.parent, "instance", None)

    def __call__(self, value: List[str]):
        qs = Applicatie.objects.all()

        if self.instance:
            qs = qs.exclude(id=self.instance.id)

        existing = qs.filter(client_ids__overlap=value).values_list(
            "uuid", "client_ids"
        )
        if existing:
            client_ids = set()
            for _existing in existing:
                client_ids = client_ids.union(_existing[1])

            raise ValidationError(
                self.message.format(
                    client_id=", ".join(client_ids.intersection(set(value))),
                    app_id=", ".join([str(_existing[0]) for _existing in existing]),
                ),
                code=self.code,
            )


class AutorisatieValidator:
    message = _("This field is required if `component` is {component}")
    code = "required"

    REQUIRED_FIELDS_PER_COMPONENT = {
        ComponentTypes.zrc: ("max_vertrouwelijkheidaanduiding", "zaaktype"),
        ComponentTypes.drc: ("max_vertrouwelijkheidaanduiding", "informatieobjecttype"),
        ComponentTypes.brc: ("besluittype",),
    }
    MAIN_RESOURCES_FOR_COMPONENTS = {
        ComponentTypes.zrc: "zaken",
        ComponentTypes.drc: "documenten",
        ComponentTypes.brc: "besluiten",
    }

    def __call__(self, autorisatie: dict) -> None:
        component = autorisatie["component"]

        if component not in self.REQUIRED_FIELDS_PER_COMPONENT:
            return

        prefix = self.MAIN_RESOURCES_FOR_COMPONENTS[component]
        if not any(scope.startswith(f"{prefix}.") for scope in autorisatie["scopes"]):
            return

        error_dict = {}
        for field_name in self.REQUIRED_FIELDS_PER_COMPONENT[component]:
            if not autorisatie[field_name]:
                error_dict.update(
                    {
                        field_name: ErrorDetail(
                            self.message.format(
                                fields=underscore_to_camel(field_name),
                                component=component,
                            ),
                            self.code,
                        )
                    }
                )

        if error_dict:
            raise ValidationError(error_dict)



# Generated by Django 2.0.6 on 2018-11-05 15:27

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="JWTSecret",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "identifier",
                    models.CharField(
                        max_length=50, unique=True, verbose_name="identifier"
                    ),
                ),
                ("secret", models.CharField(max_length=255, verbose_name="secret")),
            ],
        )
    ]



import logging
from urllib.parse import urlencode, urlparse

from django.core.exceptions import ValidationError
from django.core.validators import URLValidator
from django.db import models
from django.forms.widgets import URLInput
from django.http import QueryDict
from django.utils.translation import ugettext_lazy as _

from django_filters import fields, filters
from django_filters.constants import EMPTY_VALUES
from django_filters.rest_framework import DjangoFilterBackend
from djangorestframework_camel_case.parser import CamelCaseJSONParser
from djangorestframework_camel_case.render import CamelCaseJSONRenderer
from djangorestframework_camel_case.util import underscoreize
from rest_framework.request import Request
from rest_framework.views import APIView

from .constants import FILTER_URL_DID_NOT_RESOLVE
from .search import is_search_view
from .utils import NotAViewSet, get_resource_for_path
from .validators import validate_rsin

logger = logging.getLogger(__name__)


class Backend(DjangoFilterBackend):

    # Taken from drf_yasg.inspectors.field.CamelCaseJSONFilter
    def _is_camel_case(self, view):
        return any(
            issubclass(parser, CamelCaseJSONParser) for parser in view.parser_classes
        ) or any(
            issubclass(renderer, CamelCaseJSONRenderer)
            for renderer in view.renderer_classes
        )

    def _transform_query_params(self, view, query_params: QueryDict) -> QueryDict:
        if not self._is_camel_case(view):
            return query_params

        # data can be a regular dict if it's coming from a serializer
        if hasattr(query_params, "lists"):
            data = dict(query_params.lists())
        else:
            data = query_params

        transformed = underscoreize(data)

        return QueryDict(urlencode(transformed, doseq=True))

    def get_filterset_kwargs(
        self, request: Request, queryset: models.QuerySet, view: APIView
    ):
        """
        Get the initialization parameters for the filterset.

        * filter on request.data if request.query_params is empty
        * do the camelCase transformation of filter parameters
        """
        kwargs = super().get_filterset_kwargs(request, queryset, view)
        filter_parameters = (
            request.query_params if not is_search_view(view) else request.data
        )
        query_params = self._transform_query_params(view, filter_parameters)
        kwargs["data"] = query_params
        return kwargs


class URLModelChoiceField(fields.ModelChoiceField):
    widget = URLInput

    def __init__(self, *args, **kwargs):
        self.instance_path = kwargs.pop("instance_path", None)
        super().__init__(*args, **kwargs)

    # Placeholder - gets replaced by URLModelChoiceFilter
    def _get_request(self):
        return None

    def url_to_pk(self, url: str):
        parsed = urlparse(url)
        path = parsed.path

        # this field only supports local FKs - so if we see a domain that does
        # not match the current host, this cannot possibly yield any results
        request = self._get_request()
        if request is not None:
            host = request.get_host()
            if parsed.netloc != host:
                raise NotAViewSet("External URL cannot map to a local viewset")

        instance = get_resource_for_path(path)
        if self.instance_path:
            for bit in self.instance_path.split("."):
                instance = getattr(instance, bit)
        model = self.queryset.model
        if not isinstance(instance, model):
            raise ValidationError(
                _("Invalid resource type supplied, expected %r") % model,
                code="invalid-type",
            )
        return instance.pk

    def to_python(self, value: str):
        if value is not None:
            URLValidator()(value)

        if value:
            try:
                value = self.url_to_pk(value)
            except NotAViewSet:
                logger.info("No %s found for URL %s", self.label, value)
                return FILTER_URL_DID_NOT_RESOLVE
            except models.ObjectDoesNotExist:
                logger.info("No %s found for URL %s", self.label, value)
                return FILTER_URL_DID_NOT_RESOLVE
        return super().to_python(value)


class URLModelChoiceFilter(filters.ModelChoiceFilter):
    field_class = URLModelChoiceField

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.instance_path = kwargs.get("instance_path", None)
        self.queryset = kwargs.get("queryset")

    @property
    def field(self):
        field = super().field
        # we need access to the request in the backing field...
        field._get_request = self.get_request
        return field

    def filter(self, qs, value):
        # If the URL did not resolve to an instance, return no results
        if value == FILTER_URL_DID_NOT_RESOLVE:
            return qs.none()
        return super().filter(qs, value)


class RSINFilter(filters.CharFilter):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("validators", [validate_rsin])
        super().__init__(*args, **kwargs)


class WildcardFilter(filters.CharFilter):
    """
    Filters the queryset based on a string and optionally allows wildcards in
    the query parameter.
    """

    wildcard = "*"

    def __init__(self, *args, **kwargs):
        kwargs["lookup_expr"] = "iregex"
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        if value in EMPTY_VALUES:
            return qs

        value = r"^{}$".format(value.replace(self.wildcard, ".*"))

        return super().filter(qs, value)



# Generated by Django 2.1.1 on 2019-07-12 15:41

from django.db import migrations, models

import vng_api_common.fields


class Migration(migrations.Migration):

    dependencies = [("authorizations", "0007_auto_20190506_1212")]

    operations = [
        migrations.AlterField(
            model_name="authorizationsconfig",
            name="component",
            field=models.CharField(
                choices=[
                    ("ac", "Autorisatiecomponent"),
                    ("nrc", "Notificatierouteringcomponent"),
                    ("zrc", "Zaakregistratiecomponent"),
                    ("ztc", "Zaaktypecatalogus"),
                    ("drc", "Documentregistratiecomponent"),
                    ("brc", "Besluitregistratiecomponent"),
                ],
                default="zrc",
                max_length=50,
                verbose_name="component",
            ),
        ),
        migrations.AlterField(
            model_name="autorisatie",
            name="component",
            field=models.CharField(
                choices=[
                    ("ac", "Autorisatiecomponent"),
                    ("nrc", "Notificatierouteringcomponent"),
                    ("zrc", "Zaakregistratiecomponent"),
                    ("ztc", "Zaaktypecatalogus"),
                    ("drc", "Documentregistratiecomponent"),
                    ("brc", "Besluitregistratiecomponent"),
                ],
                help_text="Component waarop autorisatie van toepassing is.",
                max_length=50,
                verbose_name="component",
            ),
        ),
        migrations.AlterField(
            model_name="autorisatie",
            name="max_vertrouwelijkheidaanduiding",
            field=vng_api_common.fields.VertrouwelijkheidsAanduidingField(
                blank=True,
                choices=[
                    ("openbaar", "Openbaar"),
                    ("beperkt_openbaar", "Beperkt openbaar"),
                    ("intern", "Intern"),
                    ("zaakvertrouwelijk", "Zaakvertrouwelijk"),
                    ("vertrouwelijk", "Vertrouwelijk"),
                    ("confidentieel", "Confidentieel"),
                    ("geheim", "Geheim"),
                    ("zeer_geheim", "Zeer geheim"),
                ],
                help_text="Maximaal toegelaten vertrouwelijkheidaanduiding (inclusief).",
                max_length=20,
            ),
        ),
    ]



from django.conf import settings
from django.utils.module_loading import import_string

from drf_yasg.utils import swagger_auto_schema
from rest_framework import status
from rest_framework.response import Response
from rest_framework.views import APIView

from ...permissions import AuthScopesRequired
from ...scopes import Scope
from ...serializers import FoutSerializer, ValidatieFoutSerializer
from ..constants import SCOPE_NOTIFICATIES_PUBLICEREN_LABEL
from .serializers import NotificatieSerializer


class NotificationBaseView(APIView):
    """
    Abstract view to receive webhooks
    """

    swagger_schema = None

    permission_classes = (AuthScopesRequired,)
    required_scopes = Scope(
        SCOPE_NOTIFICATIES_PUBLICEREN_LABEL
    )  # FIXME: this should be standalone!

    def get_serializer(self, *args, **kwargs):
        return NotificatieSerializer(*args, **kwargs)

    @swagger_auto_schema(
        responses={
            204: "",
            400: ValidatieFoutSerializer,
            401: FoutSerializer,
            403: FoutSerializer,
            429: FoutSerializer,
            500: FoutSerializer,
            502: FoutSerializer,
            503: FoutSerializer,
        },
        operation_id="notification_receive",
        tags=["Notificaties"],
    )
    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.handle_notification(serializer.validated_data)
        return Response(status=status.HTTP_204_NO_CONTENT)

    def handle_notification(self, message):
        raise NotImplementedError("You must implemented `handle_notification`")


class NotificationView(NotificationBaseView):
    action = "create"
    permission_classes = (AuthScopesRequired,)
    required_scopes = {"create": Scope(SCOPE_NOTIFICATIES_PUBLICEREN_LABEL)}

    def create(self, request, *args, **kwargs):
        return self.post(request, *args, **kwargs)

    def handle_notification(self, message: dict) -> None:
        _handler = getattr(
            settings,
            "DEFAULT_NOTIFICATIONS_HANDLER",
            "vng_api_common.notifications.handlers.default",
        )
        handler = import_string(_handler)
        handler.handle(message)



from typing import Union
from urllib.parse import urlparse

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models import ObjectDoesNotExist
from django.utils.translation import ugettext_lazy as _

from rest_framework import permissions
from rest_framework.renderers import BrowsableAPIRenderer
from rest_framework.request import Request
from rest_framework.serializers import ValidationError

from .scopes import Scope
from .utils import get_resource_for_path


def get_required_scopes(view) -> Union[Scope, None]:
    if not hasattr(view, "required_scopes"):
        raise ImproperlyConfigured(
            "The View(Set) must have a `required_scopes` attribute"
        )

    scopes_required = view.required_scopes.get(view.action)
    return scopes_required


def bypass_permissions(request: Request) -> bool:
    """
    Bypass permission checks in DBEUG when using the browsable API renderer
    """
    return settings.DEBUG and isinstance(
        request.accepted_renderer, BrowsableAPIRenderer
    )


class ClientIdRequired(permissions.BasePermission):
    """
    Look at the client_id of an object and check that it equals client_id in the JWT
    """

    def has_object_permission(self, request: Request, view, obj) -> bool:
        if bypass_permissions(request):
            return True

        if request.method in permissions.SAFE_METHODS:
            return True
        else:
            client_id = request.jwt_auth.client_id
            return client_id == obj.client_id


class BaseAuthRequired(permissions.BasePermission):
    """
    Look at the scopes required for the current action
    and check that they are present in the AC for this client
    """

    permission_fields = ()
    get_obj = None
    obj_path = None

    def _get_obj(self, view, request):
        if not isinstance(self.get_obj, str):
            raise TypeError(
                "'get_obj' must be set to a string, representing a view method name"
            )

        method = getattr(view, self.get_obj)
        return method()

    def _get_obj_from_path(self, obj):
        if not isinstance(self.obj_path, str):
            raise TypeError(
                "'obj_path' must be a python dotted path to the main object FK"
            )

        bits = self.obj_path.split(".")
        for bit in bits:
            obj = getattr(obj, bit)
        return obj

    def _extract_field_value(self, main_obj, field):
        return getattr(main_obj, field)

    def has_permission(self, request: Request, view) -> bool:
        from rest_framework.viewsets import ViewSetMixin

        if bypass_permissions(request):
            return True

        scopes_required = get_required_scopes(view)

        if view.action == "create":
            try:
                main_obj = self._get_obj(view, request)
            except ObjectDoesNotExist:
                raise ValidationError(
                    {
                        # using self.obj_path here ASSUMES that the same serializer is used
                        # for input as output
                        self.obj_path: ValidationError(
                            _("The object does not exist in the database"),
                            code="object-does-not-exist",
                        ).detail
                    }
                )
            fields = {
                k: self._extract_field_value(main_obj, k)
                for k in self.permission_fields
            }
            return request.jwt_auth.has_auth(scopes_required, **fields)

        # detect if this is an unsupported method - if it's a viewset and the
        # action was not mapped, it's not supported and DRF will catch it
        if isinstance(view, ViewSetMixin) and view.action is None:
            return True

        # by default - check if the action is allowed at all
        return request.jwt_auth.has_auth(scopes_required)

    def has_object_permission(self, request: Request, view, obj) -> bool:
        if bypass_permissions(request):
            return True

        scopes_required = get_required_scopes(view)
        main_obj = self._get_obj_from_path(obj)
        fields = {k: getattr(main_obj, k) for k in self.permission_fields}

        return request.jwt_auth.has_auth(scopes_required, **fields)


class AuthScopesRequired(BaseAuthRequired):
    def _get_obj(self, view, request):
        return None

    def _get_obj_from_path(self, obj):
        return None


class MainObjAuthScopesRequired(BaseAuthRequired):
    def _get_obj(self, view, request):
        return request.data

    def _get_obj_from_path(self, obj):
        return obj

    def _extract_field_value(self, main_obj, field):
        return main_obj.get(field, None)


class RelatedObjAuthScopesRequired(BaseAuthRequired):
    def _get_obj(self, view, request):
        main_obj_str = request.data.get(self.obj_path, None)
        main_obj_url = urlparse(main_obj_str).path
        main_obj = get_resource_for_path(main_obj_url)
        return main_obj


def permission_class_factory(base=BaseAuthRequired, **attrs) -> type:
    """
    Build a view-specific permission class

    This is just a small wrapper around ``type`` intended to keep the code readable.
    """
    name = base.__name__
    return type(name, (base,), attrs)



# Generated by Django 2.0.13 on 2019-04-17 11:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("vng_api_common", "0002_apicredential")]

    operations = [
        migrations.AlterModelOptions(
            name="apicredential",
            options={
                "verbose_name": "external API credential",
                "verbose_name_plural": "external API credentials",
            },
        ),
        migrations.AlterModelOptions(
            name="jwtsecret",
            options={
                "verbose_name": "client credential",
                "verbose_name_plural": "client credentials",
            },
        ),
        migrations.AddField(
            model_name="apicredential",
            name="label",
            field=models.CharField(
                default="",
                help_text="human readable label",
                max_length=100,
                verbose_name="label",
            ),
        ),
    ]



import inspect
import logging
from collections import OrderedDict

from django.apps import apps
from django.conf import settings
from django.utils.translation import ugettext, ugettext_lazy as _

from drf_yasg import openapi
from drf_yasg.inspectors import SwaggerAutoSchema
from rest_framework import exceptions, serializers, status, viewsets

from ..constants import HEADER_APPLICATION, HEADER_AUDIT, HEADER_USER_ID, VERSION_HEADER
from ..exceptions import Conflict, Gone, PreconditionFailed
from ..geo import GeoMixin
from ..permissions import BaseAuthRequired, get_required_scopes
from ..search import is_search_view
from ..serializers import FoutSerializer, ValidatieFoutSerializer

logger = logging.getLogger(__name__)

TYPE_TO_FIELDMAPPING = {
    openapi.TYPE_INTEGER: serializers.IntegerField,
    openapi.TYPE_NUMBER: serializers.FloatField,
    openapi.TYPE_STRING: serializers.CharField,
    openapi.TYPE_BOOLEAN: serializers.BooleanField,
}

COMMON_ERRORS = [
    exceptions.AuthenticationFailed,
    exceptions.NotAuthenticated,
    exceptions.PermissionDenied,
    exceptions.NotAcceptable,
    Conflict,
    Gone,
    exceptions.UnsupportedMediaType,
    exceptions.Throttled,
    exceptions.APIException,
]

DEFAULT_ACTION_ERRORS = {
    "create": COMMON_ERRORS + [exceptions.ParseError, exceptions.ValidationError],
    "list": COMMON_ERRORS,
    "retrieve": COMMON_ERRORS + [exceptions.NotFound],
    "update": COMMON_ERRORS
    + [exceptions.ParseError, exceptions.ValidationError, exceptions.NotFound],
    "partial_update": COMMON_ERRORS
    + [exceptions.ParseError, exceptions.ValidationError, exceptions.NotFound],
    "destroy": COMMON_ERRORS + [exceptions.NotFound],
}

HTTP_STATUS_CODE_TITLES = {
    status.HTTP_100_CONTINUE: "Continue",
    status.HTTP_101_SWITCHING_PROTOCOLS: "Switching protocols",
    status.HTTP_200_OK: "OK",
    status.HTTP_201_CREATED: "Created",
    status.HTTP_202_ACCEPTED: "Accepted",
    status.HTTP_203_NON_AUTHORITATIVE_INFORMATION: "Non authoritative information",
    status.HTTP_204_NO_CONTENT: "No content",
    status.HTTP_205_RESET_CONTENT: "Reset content",
    status.HTTP_206_PARTIAL_CONTENT: "Partial content",
    status.HTTP_207_MULTI_STATUS: "Multi status",
    status.HTTP_300_MULTIPLE_CHOICES: "Multiple choices",
    status.HTTP_301_MOVED_PERMANENTLY: "Moved permanently",
    status.HTTP_302_FOUND: "Found",
    status.HTTP_303_SEE_OTHER: "See other",
    status.HTTP_304_NOT_MODIFIED: "Not modified",
    status.HTTP_305_USE_PROXY: "Use proxy",
    status.HTTP_306_RESERVED: "Reserved",
    status.HTTP_307_TEMPORARY_REDIRECT: "Temporary redirect",
    status.HTTP_400_BAD_REQUEST: "Bad request",
    status.HTTP_401_UNAUTHORIZED: "Unauthorized",
    status.HTTP_402_PAYMENT_REQUIRED: "Payment required",
    status.HTTP_403_FORBIDDEN: "Forbidden",
    status.HTTP_404_NOT_FOUND: "Not found",
    status.HTTP_405_METHOD_NOT_ALLOWED: "Method not allowed",
    status.HTTP_406_NOT_ACCEPTABLE: "Not acceptable",
    status.HTTP_407_PROXY_AUTHENTICATION_REQUIRED: "Proxy authentication required",
    status.HTTP_408_REQUEST_TIMEOUT: "Request timeout",
    status.HTTP_409_CONFLICT: "Conflict",
    status.HTTP_410_GONE: "Gone",
    status.HTTP_411_LENGTH_REQUIRED: "Length required",
    status.HTTP_412_PRECONDITION_FAILED: "Precondition failed",
    status.HTTP_413_REQUEST_ENTITY_TOO_LARGE: "Request entity too large",
    status.HTTP_414_REQUEST_URI_TOO_LONG: "Request uri too long",
    status.HTTP_415_UNSUPPORTED_MEDIA_TYPE: "Unsupported media type",
    status.HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE: "Requested range not satisfiable",
    status.HTTP_417_EXPECTATION_FAILED: "Expectation failed",
    status.HTTP_422_UNPROCESSABLE_ENTITY: "Unprocessable entity",
    status.HTTP_423_LOCKED: "Locked",
    status.HTTP_424_FAILED_DEPENDENCY: "Failed dependency",
    status.HTTP_428_PRECONDITION_REQUIRED: "Precondition required",
    status.HTTP_429_TOO_MANY_REQUESTS: "Too many requests",
    status.HTTP_431_REQUEST_HEADER_FIELDS_TOO_LARGE: "Request header fields too large",
    status.HTTP_451_UNAVAILABLE_FOR_LEGAL_REASONS: "Unavailable for legal reasons",
    status.HTTP_500_INTERNAL_SERVER_ERROR: "Internal server error",
    status.HTTP_501_NOT_IMPLEMENTED: "Not implemented",
    status.HTTP_502_BAD_GATEWAY: "Bad gateway",
    status.HTTP_503_SERVICE_UNAVAILABLE: "Service unavailable",
    status.HTTP_504_GATEWAY_TIMEOUT: "Gateway timeout",
    status.HTTP_505_HTTP_VERSION_NOT_SUPPORTED: "HTTP version not supported",
    status.HTTP_507_INSUFFICIENT_STORAGE: "Insufficient storage",
    status.HTTP_511_NETWORK_AUTHENTICATION_REQUIRED: "Network authentication required",
}

AUDIT_TRAIL_ENABLED = apps.is_installed("vng_api_common.audittrails")

AUDIT_REQUEST_HEADERS = [
    openapi.Parameter(
        name=HEADER_APPLICATION,
        type=openapi.TYPE_STRING,
        in_=openapi.IN_HEADER,
        required=False,
        description=ugettext("Application that performs request"),
    ),
    openapi.Parameter(
        name=HEADER_USER_ID,
        type=openapi.TYPE_STRING,
        in_=openapi.IN_HEADER,
        required=False,
        description=ugettext("Identifier of the user that performs request"),
    ),
    openapi.Parameter(
        name=HEADER_AUDIT,
        type=openapi.TYPE_STRING,
        in_=openapi.IN_HEADER,
        required=False,
        description=ugettext("Explanation why the request is done"),
    ),
]


def response_header(description: str, type: str, format: str = None) -> OrderedDict:
    header = OrderedDict(
        (("schema", OrderedDict((("type", type),))), ("description", description))
    )
    if format is not None:
        header["schema"]["format"] = format
    return header


version_header = response_header(
    "Geeft een specifieke API-versie aan in de context van een specifieke aanroep. Voorbeeld: 1.2.1.",
    type=openapi.TYPE_STRING,
)

location_header = response_header(
    "URL waar de resource leeft.", type=openapi.TYPE_STRING, format=openapi.FORMAT_URI
)


def _view_supports_audittrail(view: viewsets.ViewSet) -> bool:
    if not AUDIT_TRAIL_ENABLED:
        return False

    if not hasattr(view, "action"):
        logger.debug("Could not determine view action for view %r", view)
        return False

    # local imports, since you get errors if you try to import non-installed app
    # models
    from ..audittrails.viewsets import AuditTrailMixin

    relevant_bases = [
        base for base in view.__class__.__bases__ if issubclass(base, AuditTrailMixin)
    ]
    if not relevant_bases:
        return False

    # check if the view action is listed in any of the audit trail mixins
    action = view.action
    if action == "partial_update":  # partial update is self.update(partial=True)
        action = "update"

    # if the current view action is not provided by any of the audit trail
    # related bases, then it's not audit trail enabled
    action_in_audit_bases = any(
        action in dict(inspect.getmembers(base)) for base in relevant_bases
    )

    return action_in_audit_bases


class AutoSchema(SwaggerAutoSchema):
    @property
    def model(self):
        if hasattr(self.view, "get_queryset"):
            qs = self.view.get_queryset()
            return qs.model
        return None

    @property
    def _is_search_view(self):
        return is_search_view(self.view)

    def get_operation_id(self, operation_keys=None) -> str:
        """
        Simply return the model name as lowercase string, postfixed with the operation name.
        """
        operation_keys = operation_keys or self.operation_keys

        operation_id = self.overrides.get("operation_id", "")
        if operation_id:
            return operation_id

        action = operation_keys[-1]
        if self.model is not None:
            model_name = self.model._meta.model_name
            return f"{model_name}_{action}"
        else:
            operation_id = "_".join(operation_keys)
            return operation_id

    def should_page(self):
        if self._is_search_view:
            return hasattr(self.view, "paginator")
        return super().should_page()

    def get_request_serializer(self):
        if not self._is_search_view:
            return super().get_request_serializer()

        Base = self.view.get_search_input_serializer_class()

        filter_fields = []
        for filter_backend in self.view.filter_backends:
            filter_fields += (
                self.probe_inspectors(
                    self.filter_inspectors, "get_filter_parameters", filter_backend()
                )
                or []
            )

        filters = {}
        for filter_field in filter_fields:
            FieldClass = TYPE_TO_FIELDMAPPING[filter_field.type]
            filters[filter_field.name] = FieldClass(
                help_text=filter_field.description, required=filter_field.required
            )

        SearchSerializer = type(Base.__name__, (Base,), filters)
        return SearchSerializer()

    def _get_search_responses(self):
        response_status = status.HTTP_200_OK
        response_schema = self.serializer_to_schema(self.get_view_serializer())
        schema = openapi.Schema(type=openapi.TYPE_ARRAY, items=response_schema)
        if self.should_page():
            schema = self.get_paginated_response(schema) or schema
        return OrderedDict({str(response_status): schema})

    def _get_error_responses(self) -> OrderedDict:
        """
        Add the appropriate possible error responses to the schema.

        E.g. - we know that HTTP 400 on a POST/PATCH/PUT leads to validation
        errors, 403 to Permission Denied etc.
        """
        responses = {}

        action = self.view.action

        if (
            action not in DEFAULT_ACTION_ERRORS and self._is_search_view
        ):  # similar to a CREATE
            action = "create"

        exception_klasses = DEFAULT_ACTION_ERRORS.get(action)
        if exception_klasses is None:
            logger.debug("Unknown action %s, no default error responses added")
            return responses

        fout_schema = self.serializer_to_schema(FoutSerializer())
        for exception_klass in exception_klasses:
            status_code = exception_klass.status_code
            responses[status_code] = fout_schema

        has_validation_errors = self.get_filter_parameters() or any(
            issubclass(klass, exceptions.ValidationError) for klass in exception_klasses
        )
        if has_validation_errors:
            schema = self.serializer_to_schema(ValidatieFoutSerializer())
            responses[exceptions.ValidationError.status_code] = schema

        if isinstance(self.view, GeoMixin):
            status_code = PreconditionFailed.status_code
            responses[status_code] = fout_schema

        # sort by status code

        return OrderedDict(
            [
                (
                    status_code,
                    openapi.Response(
                        description=HTTP_STATUS_CODE_TITLES.get(status_code, ""),
                        schema=schema,
                    ),
                )
                for status_code, schema in sorted(responses.items())
            ]
        )

    def get_default_responses(self) -> OrderedDict:
        if self._is_search_view:
            responses = self._get_search_responses()
            serializer = self.get_view_serializer()
        else:
            responses = super().get_default_responses()
            serializer = self.get_request_serializer() or self.get_view_serializer()

        # inject any headers
        _responses = OrderedDict()
        for status_, schema in responses.items():
            custom_headers = (
                self.probe_inspectors(
                    self.field_inspectors,
                    "get_response_headers",
                    serializer,
                    {"field_inspectors": self.field_inspectors},
                    status=status_,
                )
                or None
            )

            assert isinstance(schema, openapi.Schema.OR_REF) or schema == ""
            response = openapi.Response(
                description=HTTP_STATUS_CODE_TITLES.get(int(status_), ""),
                schema=schema or None,
                headers=custom_headers,
            )
            _responses[status_] = response

        for status_code, response in self._get_error_responses().items():
            _responses[status_code] = response

        return _responses

    def get_response_schemas(self, response_serializers):
        responses = super().get_response_schemas(response_serializers)

        # add the Api-Version headers
        for status_code, response in responses.items():
            response.setdefault("headers", OrderedDict())
            response["headers"][VERSION_HEADER] = version_header

            if status_code == "201":
                response["headers"]["Location"] = location_header

        return responses

    def add_manual_parameters(self, parameters):
        base = super().add_manual_parameters(parameters)
        if self._is_search_view:
            serializer = self.get_request_serializer()
        else:
            serializer = self.get_request_serializer() or self.get_view_serializer()
        extra = (
            self.probe_inspectors(
                self.field_inspectors,
                "get_request_header_parameters",
                serializer,
                {"field_inspectors": self.field_inspectors},
            )
            or []
        )
        result = base + extra

        if _view_supports_audittrail(self.view):
            result += AUDIT_REQUEST_HEADERS

        return result

    def get_security(self):
        """Return a list of security requirements for this operation.

        Returning an empty list marks the endpoint as unauthenticated (i.e. removes all accepted
        authentication schemes). Returning ``None`` will inherit the top-level secuirty requirements.

        :return: security requirements
        :rtype: list[dict[str,list[str]]]"""
        permissions = self.view.get_permissions()
        scope_permissions = [
            perm for perm in permissions if isinstance(perm, BaseAuthRequired)
        ]

        if not scope_permissions:
            return super().get_security()

        if len(permissions) != len(scope_permissions):
            logger.warning(
                "Can't represent all permissions in OAS for path %s and method %s",
                self.path,
                self.method,
            )

        required_scopes = []
        for perm in scope_permissions:
            scopes = get_required_scopes(self.view)
            if scopes is None:
                continue
            required_scopes.append(scopes)

        if not required_scopes:
            return None  # use global security

        scopes = [str(scope) for scope in sorted(required_scopes)]

        # operation level security
        return [{settings.SECURITY_DEFINITION_NAME: scopes}]

    # patch around drf-yasg not taking overrides into account
    # TODO: contribute back in PR
    def get_produces(self) -> list:
        produces = super().get_produces()
        return self.overrides.get("produces", produces)


# translations aren't picked up/defined in DRF, so we need to hook them up here
_("A page number within the paginated result set.")
_("Number of results to return per page.")



import logging

from django.db import transaction
from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers

from ..constants import ComponentTypes
from ..polymorphism import Discriminator, PolymorphicSerializer
from ..serializers import add_choice_values_help_text
from .models import Applicatie, Autorisatie
from .validators import AutorisatieValidator, UniqueClientIDValidator

logger = logging.getLogger(__name__)


class AutorisatieBaseSerializer(PolymorphicSerializer):
    discriminator = Discriminator(
        discriminator_field="component",
        mapping={
            ComponentTypes.zrc: ("zaaktype", "max_vertrouwelijkheidaanduiding"),
            ComponentTypes.drc: (
                "informatieobjecttype",
                "max_vertrouwelijkheidaanduiding",
            ),
            ComponentTypes.brc: ("besluittype",),
            ComponentTypes.brc: ("besluittype",),
            ComponentTypes.nrc: (),
            ComponentTypes.ztc: (),
            ComponentTypes.ac: (),
        },
    )

    component_weergave = serializers.CharField(
        source="get_component_display",
        read_only=True,
        help_text=_("Omschrijving van `component`."),
    )

    class Meta:
        model = Autorisatie
        fields = ("component", "component_weergave", "scopes")
        extra_kwargs = {
            "scopes": {
                "help_text": _(
                    "Lijst van scope labels. Elke scope geeft toegang tot een "
                    "set van acties/operaties, zoals gedocumenteerd bij de "
                    "betreffende component."
                )
            }
        }
        validators = [AutorisatieValidator()]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        value_display_mapping = add_choice_values_help_text(ComponentTypes)
        self.fields["component"].help_text += f"\n\n{value_display_mapping}"


class ApplicatieSerializer(serializers.HyperlinkedModelSerializer):
    autorisaties = AutorisatieBaseSerializer(many=True, required=False)

    class Meta:
        model = Applicatie
        fields = (
            "url",
            "client_ids",
            "label",
            "heeft_alle_autorisaties",
            "autorisaties",
        )
        extra_kwargs = {
            "url": {"lookup_field": "uuid"},
            "heeft_alle_autorisaties": {"required": False},
            "client_ids": {
                "validators": [UniqueClientIDValidator()],
                "help_text": _(
                    "Lijst van consumer identifiers (hun 'client_id'). Een "
                    "`client_id` mag slechts bij één applicatie-object voorkomen."
                ),
            },
        }

    def validate(self, attrs):
        validated_attrs = super().validate(attrs)

        # either autorisaties or heeft_alle_autorisaties can be specified
        autorisaties_obj = None
        heeft_alle_autorisaties_obj = None
        # in case of update:
        if self.instance:
            autorisaties_obj = self.instance.autorisaties.all()
            heeft_alle_autorisaties_obj = self.instance.heeft_alle_autorisaties

        autorisaties = validated_attrs.get("autorisaties", autorisaties_obj)
        heeft_alle_autorisaties = validated_attrs.get(
            "heeft_alle_autorisaties", heeft_alle_autorisaties_obj
        )

        if autorisaties and heeft_alle_autorisaties is True:
            raise serializers.ValidationError(
                _("Either autorisaties or heeft_alle_autorisaties can be specified"),
                code="ambiguous-authorizations-specified",
            )

        if not autorisaties and heeft_alle_autorisaties is not True:
            raise serializers.ValidationError(
                _("Either autorisaties or heeft_alle_autorisaties should be specified"),
                code="missing-authorizations",
            )

        return validated_attrs

    @transaction.atomic
    def create(self, validated_data):
        autorisaties_data = validated_data.pop("autorisaties", None)
        applicatie = super().create(validated_data)

        if autorisaties_data:
            for auth in autorisaties_data:
                Autorisatie.objects.create(**auth, applicatie=applicatie)

        return applicatie

    @transaction.atomic
    def update(self, instance, validated_data):
        autorisaties_data = validated_data.pop("autorisaties", None)
        applicatie = super().update(instance, validated_data)

        # in case of update autorisaties - remove all related autorisaties
        if autorisaties_data is not None:
            applicatie.autorisaties.all().delete()
            for auth in autorisaties_data:
                Autorisatie.objects.create(**auth, applicatie=applicatie)

        return applicatie


class ApplicatieUuidSerializer(ApplicatieSerializer):
    """
    Serializer for saving data in local auth DB
    uuid is used for synchronizing identifiers with AC DB
    """

    class Meta(ApplicatieSerializer.Meta):
        fields = ApplicatieSerializer.Meta.fields + ("uuid",)



from django.contrib import admin
from django.urls import include, path, re_path
from django.views.generic import RedirectView

from .schema import SchemaView
from .views import NotificationView

urlpatterns = [
    path("admin/", admin.site.urls),
    path(
        "api/",
        include(
            [
                # API documentation
                re_path(
                    r"^schema/openapi(?P<format>\.json|\.yaml)$",
                    SchemaView.without_ui(cache_timeout=None),
                    name="schema-json",
                ),
                re_path(
                    r"^schema/$",
                    SchemaView.with_ui("redoc", cache_timeout=None),
                    name="schema-redoc",
                ),
            ]
        ),
    ),
    path("api/", include("vng_api_common.api.urls")),
    path("ref/", include("vng_api_common.urls")),
    # this is a hack to get the parameter to show up in the API spec
    # this effectively makes this a wildcard URL, so it should be LAST
    path("<webhooks_path>", NotificationView.as_view()),
    path("", RedirectView.as_view(url="/api/")),
]



import logging
import re
import uuid
from typing import Union

from django.apps import apps
from django.conf import settings
from django.db import models
from django.http import HttpRequest
from django.urls import Resolver404, get_resolver
from django.utils.module_loading import import_string

from zds_client.client import ClientError

try:
    from djangorestframework_camel_case.util import (
        underscore_to_camel as _underscore_to_camel,
    )
except ImportError:
    from djangorestframework_camel_case.util import (
        underscoreToCamel as _underscore_to_camel,
    )


logger = logging.getLogger(__name__)

RE_UNDERSCORE = re.compile(r"[a-z]_[a-z]")


def get_subclasses(cls):
    for subclass in cls.__subclasses__():
        yield from get_subclasses(subclass)
        yield subclass


def lookup_kwargs_to_filters(lookup_kwargs: dict, kwargs: dict) -> dict:
    """
    Using the lookup_kwargs map and the view kwargs, construct the queryset
    filter dict.
    """
    filters = {}
    for kwarg, field_name in lookup_kwargs.items():
        if kwarg not in kwargs:
            continue
        filters[field_name] = kwargs[kwarg]
    return filters


class NotAViewSet(Exception):
    pass


def get_viewset_for_path(path: str, method="GET") -> "rest_framework.viewsets.ViewSet":
    """
    Look up which viewset matches a path.
    """
    # NOTE: this doesn't support setting a different urlconf on the request
    resolver = get_resolver()

    try:
        resolver_match = resolver.resolve(path)
    except Resolver404 as exc:
        raise models.ObjectDoesNotExist("URL did not resolve") from exc
    callback, callback_args, callback_kwargs = resolver_match

    if not hasattr(callback, "cls"):
        raise NotAViewSet(f"Callback for {path} does not look like a viewset")

    viewset = callback.cls(**callback.initkwargs)
    viewset.action_map = callback.actions
    viewset.request = HttpRequest()
    viewset.args = callback_args
    viewset.kwargs = callback_kwargs

    viewset.action = viewset.action_map.get(method.lower())

    return viewset


def get_resource_for_path(path: str) -> models.Model:
    """
    Retrieve the API instance belonging to a (detail) path.
    """
    if settings.FORCE_SCRIPT_NAME and path.startswith(settings.FORCE_SCRIPT_NAME):
        prefix_length = len(settings.FORCE_SCRIPT_NAME)
        path = path[prefix_length:]

    viewset = get_viewset_for_path(path)

    # See rest_framework.mixins.RetieveModelMixin.get_object()
    lookup_url_kwarg = viewset.lookup_url_kwarg or viewset.lookup_field
    filter_kwargs = {viewset.lookup_field: viewset.kwargs[lookup_url_kwarg]}

    return viewset.get_queryset().get(**filter_kwargs)


def underscore_to_camel(input_: Union[str, int]) -> str:
    """
    Convert a string from under_score to camelCase.
    """
    if not isinstance(input_, str):
        return input_

    return re.sub(RE_UNDERSCORE, _underscore_to_camel, input_)


def get_uuid_from_path(path: str) -> str:
    """
    Get the last path of path
    """
    if path.endswith("/"):
        path = path.rstrip("/")

    uuid_str = path.rsplit("/")[-1]

    # validate if it's a proper hex string
    uuid.UUID(uuid_str)

    return uuid_str


def request_object_attribute(
    url: str, attribute: str, resource: Union[str, None] = None
) -> str:
    from .models import APICredential

    Client = import_string(settings.ZDS_CLIENT_CLASS)
    client = Client.from_url(url)
    client.auth = APICredential.get_auth(url)
    try:
        result = client.retrieve(resource, url=url)[attribute]
    except (ClientError, KeyError) as exc:
        logger.warning(
            "%s was retrieved from %s with the %s: %s",
            attribute,
            url,
            exc.__class__.__name__,
            exc,
        )
        result = ""
    return result


def generate_unique_identification(instance: models.Model, date_field_name: str):
    model = type(instance)
    model_name = getattr(model, "IDENTIFICATIE_PREFIX", model._meta.model_name.upper())

    year = getattr(instance, date_field_name).year
    prefix = f"{model_name}-{year}"

    pattern = prefix + r"-\d{10}"

    issued_ids_for_year = model._default_manager.filter(identificatie__regex=pattern)

    if issued_ids_for_year.exists():
        max_id = issued_ids_for_year.aggregate(models.Max("identificatie"))[
            "identificatie__max"
        ]
        number = int(max_id.split("-")[-1]) + 1
    else:
        number = 1

    padded_number = str(number).zfill(10)
    return f"{prefix}-{padded_number}"


def get_help_text(model_string: str, field_name: str) -> str:
    ModelClass = apps.get_model(model_string, require_ready=False)
    field = ModelClass._meta.get_field(field_name)
    return field.help_text



from .auth import (
    AuthCheckMixin,
    JWTAuthMixin,
    JWTScopesMixin,
    generate_jwt,
    generate_jwt_auth,
)
from .schema import TypeCheckMixin, get_operation_url, get_validation_errors
from .urls import reverse, reverse_lazy

__all__ = [
    "generate_jwt",
    "AuthCheckMixin",
    "JWTScopesMixin",
    "get_operation_url",
    "TypeCheckMixin",
    "get_validation_errors",
    "reverse",
    "reverse_lazy",
    "JWTAuthMixin",
    "generate_jwt_auth",
]



from ...models import JWTSecret
from ..kanalen import Kanaal
from ..utils import notification_documentation


def test_generate_docs():
    kanaal = Kanaal(label="dummy", main_resource=JWTSecret, kenmerken=("identifier",))

    result = notification_documentation(kanaal)

    expected = """### Notificaties

Deze API publiceert notificaties op het kanaal `dummy`.

**Main resource**

`jwtsecret`



**Kenmerken**

* `identifier`: Client ID to identify external API's and applications that access this API.

**Resources en acties**
"""
    assert result == expected



import json
import logging
import os
from urllib.parse import urlsplit

from django.conf import settings

from drf_yasg import openapi
from drf_yasg.app_settings import swagger_settings
from drf_yasg.codecs import yaml_sane_dump, yaml_sane_load
from drf_yasg.generators import OpenAPISchemaGenerator as _OpenAPISchemaGenerator
from drf_yasg.renderers import SwaggerJSONRenderer, SwaggerYAMLRenderer, _SpecRenderer
from drf_yasg.views import get_schema_view
from rest_framework import permissions
from rest_framework.response import Response

logger = logging.getLogger(__name__)


class OpenAPISchemaGenerator(_OpenAPISchemaGenerator):
    def get_path_parameters(self, path, view_cls):
        """Return a list of Parameter instances corresponding to any templated path variables.

        :param str path: templated request path
        :param type view_cls: the view class associated with the path
        :return: path parameters
        :rtype: list[openapi.Parameter]
        """
        parameters = super().get_path_parameters(path, view_cls)

        # see if we can specify UUID a bit more
        for parameter in parameters:
            # the most pragmatic of checks
            if not parameter.name.endswith("_uuid"):
                continue
            parameter.format = openapi.FORMAT_UUID
            parameter.description = "Unieke resource identifier (UUID4)"
        return parameters


DefaultSchemaView = get_schema_view(
    # validators=['flex', 'ssv'],
    generator_class=OpenAPISchemaGenerator,
    public=True,
    permission_classes=(permissions.AllowAny,),
)


class OpenAPIV3RendererMixin:
    def render(self, data, media_type=None, renderer_context=None):
        if "openapi" in data or "swagger" in data:
            if self.format == ".yaml":
                return yaml_sane_dump(data, False)
            elif self.format == ".json":
                return json.dumps(data)

        return super().render(
            data, media_type=media_type, renderer_context=renderer_context
        )


SPEC_RENDERERS = (
    type("SwaggerYAMLRenderer", (OpenAPIV3RendererMixin, SwaggerYAMLRenderer), {}),
    type("SwaggerJSONRenderer", (OpenAPIV3RendererMixin, SwaggerJSONRenderer), {}),
)


class SchemaView(DefaultSchemaView):
    """
    Always serve the v3 version, which is kept in version control.

    .. warn:: there is a risk of the generated schema not being in sync with
      the code. Unfortunately, that's the tradeoff we have. We could set up
      CI to check for outdated schemas.
    """

    schema_path = None

    @property
    def _is_openapi_v2(self) -> bool:
        default = "3" if "format" in self.kwargs else "2"
        version = self.request.GET.get("v", default)
        return version.startswith("2")

    def get_renderers(self):
        if self._is_openapi_v2:
            return super().get_renderers()
        return [renderer() for renderer in SPEC_RENDERERS]

    def get_schema_path(self) -> str:
        return self.schema_path or os.path.join(
            settings.BASE_DIR, "src", "openapi.yaml"
        )

    def get(self, request, version="", *args, **kwargs):
        if self._is_openapi_v2:
            version = request.version or version or ""
            if isinstance(request.accepted_renderer, _SpecRenderer):
                generator = self.generator_class(
                    getattr(self, "info", swagger_settings.DEFAULT_INFO),
                    version,
                    None,
                    patterns,
                    urlconf,
                )
            else:
                generator = self.generator_class(
                    getattr(self, "info", swagger_settings.DEFAULT_INFO),
                    version,
                    None,
                    patterns=[],
                )

            schema = generator.get_schema(request, self.public)
            if schema is None:
                raise exceptions.PermissionDenied()  # pragma: no cover
            return Response(schema)

        # serve the staticically included V3 schema
        SCHEMA_PATH = self.get_schema_path()
        with open(SCHEMA_PATH, "r") as infile:
            schema = yaml_sane_load(infile)

        # fix the servers
        for server in schema["servers"]:
            split_url = urlsplit(server["url"])
            if split_url.netloc:
                continue

            prefix = settings.FORCE_SCRIPT_NAME or ""
            if prefix.endswith("/"):
                prefix = prefix[:-1]
            server_path = f"{prefix}{server['url']}"
            server["url"] = request.build_absolute_uri(server_path)

        return Response(data=schema, headers={"X-OAS-Version": schema["openapi"]})



# Generated by Django 2.1.8 on 2019-05-27 05:55

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Person",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=50, verbose_name="name")),
                (
                    "address_street",
                    models.CharField(max_length=255, verbose_name="street name"),
                ),
                (
                    "address_number",
                    models.CharField(max_length=10, verbose_name="house number"),
                ),
            ],
        )
    ]



# https://pyjwt.readthedocs.io/en/latest/usage.html#reading-headers-without-validation
# -> we can put the organization/service in the headers itself
import logging
from typing import List, Optional, Union

from django.conf import settings
from django.db import models, transaction
from django.db.models import QuerySet
from django.utils.translation import ugettext as _

import jwt
from djangorestframework_camel_case.util import underscoreize
from rest_framework.exceptions import PermissionDenied
from rest_framework.response import Response
from zds_client.client import ClientError

from .authorizations.models import Applicatie, AuthorizationsConfig, Autorisatie
from .authorizations.serializers import ApplicatieUuidSerializer
from .constants import VERSION_HEADER, VertrouwelijkheidsAanduiding
from .models import JWTSecret
from .utils import get_uuid_from_path

logger = logging.getLogger(__name__)


class JWTAuth:
    def __init__(self, encoded: str = None):
        self.encoded = encoded

    @property
    def applicaties(self) -> Union[list, None]:
        if self.client_id is None:
            return []

        applicaties = self._get_auth()

        if not applicaties:
            auth_data = self._request_auth()
            applicaties = self._save_auth(auth_data)

        return applicaties

    @property
    def autorisaties(self) -> models.QuerySet:
        """
        Retrieve all authorizations relevant to this component.
        """
        app_ids = [app.id for app in self.applicaties]
        config = AuthorizationsConfig.get_solo()
        return Autorisatie.objects.filter(
            applicatie_id__in=app_ids, component=config.component
        )

    def _request_auth(self) -> list:
        client = AuthorizationsConfig.get_client()
        try:
            response = client.list(
                "applicatie", query_params={"clientIds": self.client_id}
            )
        except ClientError as exc:
            response = exc.args[0]
            # friendly debug - hint at where the problem is located
            if response["status"] == 403 and response["code"] == "not_authenticated":
                detail = _(
                    "Component could not authenticate against the AC - "
                    "authorizations could not be retrieved"
                )
                raise PermissionDenied(detail=detail, code="not_authenticated_for_ac")
            logger.warn("Authorization component can't be accessed")
            return []

        return underscoreize(response["results"])

    def _get_auth(self):
        return Applicatie.objects.filter(client_ids__contains=[self.client_id])

    @transaction.atomic
    def _save_auth(self, auth_data):
        applicaties = []

        for applicatie_data in auth_data:
            applicatie_serializer = ApplicatieUuidSerializer(data=applicatie_data)
            uuid = get_uuid_from_path(applicatie_data["url"])
            applicatie_data["uuid"] = uuid
            applicatie_serializer.is_valid()
            applicaties.append(applicatie_serializer.save())

        return applicaties

    @property
    def payload(self):
        if self.encoded is None:
            return None

        if not hasattr(self, "_payload"):
            # decode the JWT and validate it

            # jwt check
            try:
                payload = jwt.decode(self.encoded, verify=False)
            except jwt.DecodeError:
                logger.info("Invalid JWT encountered")
                raise PermissionDenied(
                    _(
                        "JWT could not be decoded. Possibly you made a copy-paste mistake."
                    ),
                    code="jwt-decode-error",
                )

            # get client_id
            try:
                client_id = payload["client_id"]
            except KeyError:
                try:
                    header = jwt.get_unverified_header(self.encoded)
                except jwt.DecodeError:
                    logger.info("Invalid JWT encountered")
                    raise PermissionDenied(
                        _(
                            "JWT could not be decoded. Possibly you made a copy-paste mistake."
                        ),
                        code="jwt-decode-error",
                    )
                else:
                    try:
                        client_id = header["client_identifier"]
                    except KeyError:
                        raise PermissionDenied(
                            "Client identifier is niet aanwezig in JWT",
                            code="missing-client-identifier",
                        )
                    else:
                        logger.warning(
                            _(
                                "The support of authorization old format will be terminated soon. "
                                "Please use a new format with the separate Authorization Component"
                            )
                        )

            # find client_id in DB and retrieve it's secret
            try:
                jwt_secret = JWTSecret.objects.get(identifier=client_id)
            except JWTSecret.DoesNotExist:
                raise PermissionDenied(
                    "Client identifier bestaat niet", code="invalid-client-identifier"
                )
            else:
                key = jwt_secret.secret

            # check signature of the token
            try:
                payload = jwt.decode(self.encoded, key, algorithms="HS256")
            except jwt.InvalidSignatureError:
                logger.exception("Invalid signature - possible payload tampering?")
                raise PermissionDenied(
                    "Client credentials zijn niet geldig", code="invalid-jwt-signature"
                )

            if "client_id" not in payload:
                payload["client_id"] = client_id

            self._payload = payload

        return self._payload

    @property
    def client_id(self):
        if not self.payload:
            return None

        return self.payload["client_id"]

    def filter_vertrouwelijkheidaanduiding(self, base: QuerySet, value) -> QuerySet:
        if value is None:
            return base

        order_provided = VertrouwelijkheidsAanduiding.get_choice(value).order
        order_case = VertrouwelijkheidsAanduiding.get_order_expression(
            "max_vertrouwelijkheidaanduiding"
        )

        # In this case we are filtering Autorisatie model to look for auth which meets our needs.
        # Therefore we're only considering authorizations here that have a max_vertrouwelijkheidaanduiding
        # bigger or equal than what we're checking for the object.
        # In cases when we are filtering data objects (Zaak, InformatieObject etc) it's the other way around

        return base.annotate(max_vertr=order_case).filter(max_vertr__gte=order_provided)

    def filter_default(self, base: QuerySet, name, value) -> QuerySet:
        if value is None:
            return base

        return base.filter(**{name: value})

    def has_auth(
        self, scopes: List[str], component: Optional[str] = None, **fields
    ) -> bool:
        if scopes is None:
            return False

        scopes_provided = set()
        config = AuthorizationsConfig.get_solo()
        if component is None:
            component = config.component

        for applicatie in self.applicaties:
            # allow everything
            if applicatie.heeft_alle_autorisaties is True:
                return True

            autorisaties = applicatie.autorisaties.filter(component=component)

            # filter on all additional components
            for field_name, field_value in fields.items():
                if hasattr(self, f"filter_{field_name}"):
                    autorisaties = getattr(self, f"filter_{field_name}")(
                        autorisaties, field_value
                    )
                else:
                    autorisaties = self.filter_default(
                        autorisaties, field_name, field_value
                    )

            for autorisatie in autorisaties:
                scopes_provided.update(autorisatie.scopes)

        return scopes.is_contained_in(list(scopes_provided))


class AuthMiddleware:

    header = "HTTP_AUTHORIZATION"
    auth_type = "Bearer"

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        self.extract_jwt_payload(request)
        return self.get_response(request) if self.get_response else None

    def extract_jwt_payload(self, request):
        authorization = request.META.get(self.header, "")
        prefix = f"{self.auth_type} "
        if authorization.startswith(prefix):
            # grab the actual token
            encoded = authorization[len(prefix) :]
        else:
            encoded = None

        request.jwt_auth = JWTAuth(encoded)


class APIVersionHeaderMiddleware:
    """
    Include a header specifying the API-version
    """

    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        if self.get_response is None:
            return None

        response = self.get_response(request)

        # not an API response, exit early
        if not isinstance(response, Response):
            return response

        # set the header
        response[VERSION_HEADER] = settings.API_VERSION

        return response



import os
from functools import lru_cache
from urllib.parse import urlparse

from django.conf import settings

import yaml

DEFAULT_PATH_PARAMETERS = {"version": "1"}

SPEC_PATH = os.path.join(settings.BASE_DIR, "src", "openapi.yaml")


@lru_cache()
def get_spec(path: str = SPEC_PATH) -> dict:
    with open(path, "r") as infile:
        spec = yaml.safe_load(infile)
    return spec


def get_operation_url(operation: str, spec_path: str = SPEC_PATH, **kwargs):
    """
    Look up the url of an operation from the API spec.
    """
    spec = get_spec(spec_path)
    url = spec["servers"][0]["url"]
    base_path = urlparse(url).path

    for path, methods in spec["paths"].items():
        for name, method in methods.items():
            if name == "parameters":
                continue

            if method["operationId"] == operation:
                format_kwargs = DEFAULT_PATH_PARAMETERS.copy()
                format_kwargs.update(**kwargs)
                path = path.format(**format_kwargs)
                return f"{base_path}{path}"

    raise ValueError(f"Operation {operation} not found")


class TypeCheckMixin:
    def assertResponseTypes(self, response_data: dict, types: tuple):
        """
        Do type checks on the response data.

        :param types: tuple of (field_name, class)
        :raises AssertionError: if the type mismatches
        """
        for field, type_ in types:
            with self.subTest(field=field, type=type_):
                self.assertIsInstance(response_data[field], type_)


def get_validation_errors(response, field, index=0):
    """
    Extra the validation error for ``field`` from the response.

    Assumes there's only one validation error for the field.
    """
    assert response.status_code == 400
    i = 0
    for error in response.data["invalid_params"]:
        if error["name"] != field:
            continue

        if i == index:
            return error

        i += 1



# Generated by Django 2.2.6 on 2020-04-16 04:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("testapp", "0002_auto_20190620_0849"),
    ]

    operations = [
        migrations.AddField(
            model_name="group",
            name="subgroup_field_1",
            field=models.CharField(blank=True, max_length=50),
        ),
        migrations.AddField(
            model_name="group",
            name="subgroup_field_2",
            field=models.CharField(blank=True, default="baz", max_length=50),
        ),
    ]



# Generated by Django 2.0.13 on 2019-04-17 11:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("notifications", "0005_fix_default_nrc")]

    operations = [
        migrations.RemoveField(model_name="notificationsconfig", name="client_id"),
        migrations.RemoveField(model_name="notificationsconfig", name="secret"),
        migrations.AlterField(
            model_name="notificationsconfig",
            name="api_root",
            field=models.URLField(
                default="https://ref.tst.vng.cloud/nrc/api/v1",
                unique=True,
                verbose_name="api root",
            ),
        ),
    ]



# Generated by Django 2.1.1 on 2019-08-19 12:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("audittrails", "0009_auto_20190712_1643")]

    operations = [
        migrations.AddField(
            model_name="audittrail",
            name="request_id",
            field=models.CharField(
                blank=True,
                help_text='Een globaal "request" ID om een verzoek door het netwerk heen te traceren.',
                max_length=255,
            ),
        )
    ]



import re
from urllib.parse import urlparse

from zds_client.client import UUID_PATTERN


class Response:
    def __init__(self, status_code: int = 200):
        self.status_code = status_code

    def json(self) -> dict:
        return {}


def link_fetcher_404(url: str, *args, **kwargs):
    return Response(status_code=404)


def link_fetcher_200(url: str, *args, **kwargs):
    return Response(status_code=200)


class MockClient:
    data = {}

    @classmethod
    def from_url(cls, detail_url: str):
        clients = {
            "zrc": ZRCMockClient,
            "ztc": ZTCMockClient,
            "drc": DRCMockClient,
            "brc": BRCMockClient,
            "notificaties": NotifMockClient,
        }

        parsed_url = urlparse(detail_url)

        if ":" in parsed_url.netloc:
            host, port = parsed_url.netloc.split(":")
        else:
            host = parsed_url.netloc.split(".")[0]  # Remove top-level domain from host

        # Try mock client based on host.
        if host in clients:
            return clients[host]()

        # Try mock client based first element of the path.
        first_path_element = parsed_url.path.strip("/").split("/", 1)[0]
        if first_path_element in clients:
            return clients[first_path_element]()

        # Try mock client based on last element of the path - for notifications
        last_path_element = parsed_url.path.split("/")[-1]
        if last_path_element in clients:
            return clients[last_path_element]()

        raise ValueError("Cannot determine service based on url: %s", detail_url)

    def request(self, path: str, operation: str, method="GET", **kwargs):
        bits = path.rsplit("/", 2)

        # Match UUIDs and simple numbers that are most likely used for testing.
        if re.match(r"^({}|[0-9]+)$".format(UUID_PATTERN), bits[-1]):
            resource = bits[-2]
            uuid = bits[-1]
        else:
            resource = bits[-1]
            uuid = None

        # Should the mock client really consult an OAS to figure out the
        # operation? Workaround for now is to strip off plural forms ("n", "s",
        # "sen").
        if resource not in self.data:
            if resource[0:-1] in self.data:
                resource = resource[0:-1]
            elif resource[0:-3] in self.data:
                resource = resource[0:-3]

        if method == "POST" and not uuid:
            return self.create(resource)

        if method == "GET" and not uuid:
            return self.list(resource)

        return self.retrieve(resource, uuid=uuid)

    def list(self, resource: str, *args, **kwargs):
        return self.data.get(resource, [])

    def retrieve(self, resource: str, *args, **kwargs):
        try:
            index = int(kwargs.get("uuid", 1)) - 1
        except ValueError:
            index = 0

        # result = copy.deepcopy(self.data.get(resource)[index])
        # result['url'] = result['url'].format(**kwargs)
        return self.data.get(resource)[index]

    def create(self, resource: str, *args, **kwargs):
        return self.data.get(resource, {})


class ZTCMockClient(MockClient):

    data = {
        "statustype": [
            {
                "url": "https://ztc/api/v1/catalogussen/{catalogus_uuid}/zaaktypen/{zaaktype_uuid}/statustypen/{uuid}",
                "volgnummer": 1,
                "isEindstatus": False,
            },
            {
                "url": "https://ztc/api/v1/catalogussen/{catalogus_uuid}/zaaktypen/{zaaktype_uuid}/statustypen/{uuid}",
                "volgnummer": 2,
                "isEindstatus": True,
            },
        ],
        "resultaattype": [
            {
                "url": "https://ztc/api/v1/resultaattypen/{uuid}",
                "zaaktype": "https://ztc/api/v1/catalogussen/{catalogus_uuid}/zaaktypen/{zaaktype_uuid}",
                "omschrijving": "Klaar",
                "resultaattypeomschrijving": "https://referentielijsten-api.vng.cloud/api/v1/resultaattypeomschrijvingen/e6a0c939-3404-45b0-88e3-76c94fb80ea7",
                "omschrijvingGeneriek": "Afgewezen",
                "selectielijstklasse": "https://referentielijsten-api.vng.cloud/api/v1/resultaten/d8bd516e-95b5-47ee-988d-d6624e94db1f",
                "toelichting": "",
                "archiefnominatie": "vernietigen",
                "archiefactietermijn": "P5Y",
                "brondatumArchiefprocedure": {
                    "afleidingswijze": "afgehandeld",
                    "datumkenmerk": None,
                    "einddatumBekend": False,
                    "objecttype": None,
                    "registratie": None,
                    "procestermijn": None,
                },
            }
        ],
    }


class DRCMockClient(MockClient):
    data = {
        "enkelvoudiginformatieobject": [
            {
                "url": "https://mock/enkelvoudiginformatieobjecten/1234",
                "identificatie": "9560b006-25ef-4111-9f53-966762173d41",
            }
        ]
    }


class ZRCMockClient(MockClient):
    data = {
        "zaakinformatieobject": [
            {
                "url": "https://mock/zaakinformatieobjecten/1234",
                "informatieobject": "",
                "object": "",
            }
        ],
        "zaak": [
            {
                "url": "https://zrc/api/v1/zaken/e0c464e4-727c-41ef-948d-e3109ae870f4",
                "identificatie": "3a179da6-ce9e-4723-bb8e-f47895836a9a",
            }
        ],
    }


class BRCMockClient(MockClient):
    data = {
        "besluitinformatieobject": [
            {
                "url": "https://mock/besluitinformatieobjecten/1234",
                "informatieobject": "",
                "object": "",
            }
        ],
        "besluit": [
            {
                "url": "https://brc/api/v1/besluiten/1fc80bc9-5563-448a-8fd7-46cb44207528",
                "identificatie": "3e563db0-0bcf-46f2-a881-08f76712a40d",
            }
        ],
    }


class NotifMockClient(MockClient):
    """
    for sending notifications
    """

    data = {"notificaties": {}}



from django.utils.translation import ugettext_lazy as _

from rest_framework import serializers

from ...validators import UntilNowValidator


class NotificatieSerializer(serializers.Serializer):
    kanaal = serializers.CharField(
        label=_("kanaal"),
        max_length=50,
        help_text=_(
            "De naam van het kanaal (`KANAAL.naam`) waar het bericht "
            "op moet worden gepubliceerd."
        ),
    )
    hoofd_object = serializers.URLField(
        label=_("hoofd object"),
        help_text=_(
            "URL-referentie naar het hoofd object van de publicerende "
            "API die betrekking heeft op de `resource`."
        ),
    )
    resource = serializers.CharField(
        label=_("resource"),
        max_length=100,
        help_text=_("De resourcenaam waar de notificatie over gaat."),
    )
    resource_url = serializers.URLField(
        label=_("resource URL"),
        help_text=_("URL-referentie naar de `resource` van de publicerende " "API."),
    )
    actie = serializers.CharField(
        label=_("actie"),
        max_length=100,
        help_text=_(
            "De actie die door de publicerende API is gedaan. De "
            "publicerende API specificeert de toegestane acties."
        ),
    )
    aanmaakdatum = serializers.DateTimeField(
        label=_("aanmaakdatum"),
        validators=[UntilNowValidator()],
        help_text=_("Datum en tijd waarop de actie heeft plaatsgevonden."),
    )
    kenmerken = serializers.DictField(
        label=_("kenmerken"),
        required=False,
        child=serializers.CharField(
            label=_("kenmerk"),
            max_length=1000,
            help_text=_("Een waarde behorende bij de sleutel."),
        ),
        help_text=_(
            "Mapping van kenmerken (sleutel/waarde) van de notificatie. De "
            "publicerende API specificeert de toegestane kenmerken."
        ),
    )



# Generated by Django 2.2.6 on 2019-11-14 07:28

from django.db import migrations


class Renamer:
    def __init__(self, old: str, new: str):
        self.old = old
        self.new = new

    def __call__(self, apps, schema_editor):
        Autorisatie = apps.get_model("authorizations.Autorisatie")
        for autorisatie in Autorisatie.objects.all():
            scopes = [self.patch_scope(scope) for scope in autorisatie.scopes]
            if scopes == autorisatie.scopes:
                continue

            autorisatie.scopes = scopes
            autorisatie.save(update_fields=["scopes"])

    def patch_scope(self, scope: str) -> str:
        if not scope.startswith(f"{self.old}."):
            return scope
        return scope.replace(f"{self.old}.", f"{self.new}.", 1)


class Migration(migrations.Migration):

    dependencies = [
        ("authorizations", "0010_auto_20190712_1643"),
    ]

    operations = [
        migrations.RunPython(
            Renamer("zaaktypes", "catalogi"), Renamer("catalogi", "zaaktypes"),
        )
    ]



from django.contrib import admin

from .models import AuditTrail


@admin.register(AuditTrail)
class AuditTrailAdmin(admin.ModelAdmin):
    list_display = ("uuid", "bron", "resultaat", "applicatie_weergave")
    list_filter = ("bron", "applicatie_id", "resultaat")
    date_hierarchy = "aanmaakdatum"



# Generated by Django 2.1.8 on 2019-05-06 09:01

import django.contrib.postgres.fields
import django.db.models.deletion
from django.db import migrations, models

import vng_api_common.fields


class Migration(migrations.Migration):

    dependencies = [("authorizations", "0005_auto_20190506_0842")]

    operations = [
        migrations.AlterField(
            model_name="applicatie",
            name="client_ids",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=50),
                help_text="Komma-gescheiden lijst van consumer identifiers (hun client_id).",
                size=None,
                verbose_name="client IDs",
            ),
        ),
        migrations.AlterField(
            model_name="applicatie",
            name="heeft_alle_autorisaties",
            field=models.BooleanField(
                default=False,
                help_text="Indien alle autorisaties gegeven zijn, dan hoeven deze niet individueel opgegeven te worden. Gebruik dit alleen als je de consumer helemaal vertrouwt.",
                verbose_name="heeft alle autorisaties",
            ),
        ),
        migrations.AlterField(
            model_name="applicatie",
            name="label",
            field=models.CharField(
                help_text="Een leesbare representatie van de applicatie, voor eindgebruikers.",
                max_length=100,
            ),
        ),
        migrations.AlterField(
            model_name="autorisatie",
            name="applicatie",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="autorisaties",
                to="authorizations.Applicatie",
                verbose_name="applicatie",
            ),
        ),
        migrations.AlterField(
            model_name="autorisatie",
            name="component",
            field=models.CharField(
                choices=[
                    ("AC", "Autorisatiecomponent"),
                    ("NRC", "Notificatierouteringcomponent"),
                    ("ZRC", "Zaakregistratiecomponent"),
                    ("ZTC", "Zaaktypecatalogus"),
                    ("DRC", "Documentregistratiecomponent"),
                    ("BRC", "Besluitregistratiecomponent"),
                ],
                help_text="Component waarop autorisatie van toepassing is.",
                max_length=50,
                verbose_name="component",
            ),
        ),
        migrations.AlterField(
            model_name="autorisatie",
            name="max_vertrouwelijkheidaanduiding",
            field=vng_api_common.fields.VertrouwelijkheidsAanduidingField(
                blank=True,
                choices=[
                    ("openbaar", "OPENBAAR"),
                    ("beperkt openbaar", "BEPERKT OPENBAAR"),
                    ("intern", "INTERN"),
                    ("zaakvertrouwelijk", "ZAAKVERTROUWELIJK"),
                    ("vertrouwelijk", "VERTROUWELIJK"),
                    ("confidentieel", "CONFIDENTIEEL"),
                    ("geheim", "GEHEIM"),
                    ("zeer geheim", "ZEER GEHEIM"),
                ],
                help_text="Maximaal toegelaten vertrouwelijkheidaanduiding (inclusief).",
                max_length=20,
            ),
        ),
        migrations.AlterField(
            model_name="autorisatie",
            name="scopes",
            field=django.contrib.postgres.fields.ArrayField(
                base_field=models.CharField(max_length=100),
                help_text="Komma-gescheiden lijst van scope labels.",
                size=None,
                verbose_name="scopes",
            ),
        ),
        migrations.AlterField(
            model_name="autorisatie",
            name="zaaktype",
            field=models.URLField(
                blank=True,
                help_text="URL naar het zaaktype waarop de autorisatie van toepassing is.",
                max_length=1000,
                verbose_name="zaaktype",
            ),
        ),
    ]



# Generated by Django 2.0.13 on 2019-04-17 10:17

from django.db import migrations


def fix_config(apps, _):
    NotificationsConfig = apps.get_model("notifications.NotificationsConfig")

    try:
        config = NotificationsConfig.objects.get()
    except NotificationsConfig.DoesNotExist:
        return

    if config.api_root == "https://ref.tst.vng.cloud/nc/api/v1":
        config.api_root = "https://ref.tst.vng.cloud/nrc/api/v1"
        config.save()


class Migration(migrations.Migration):

    dependencies = [("notifications", "0004_auto_20190325_1313")]

    operations = [migrations.RunPython(fix_config, migrations.RunPython.noop)]



import uuid

from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.utils.translation import ugettext_lazy as _

from ..constants import ComponentTypes, VertrouwelijkheidsAanduiding
from ..decorators import field_default
from ..fields import VertrouwelijkheidsAanduidingField
from ..models import APIMixin, ClientConfig


@field_default("api_root", "https://autorisaties-api.vng.cloud/api/v1")
class AuthorizationsConfig(ClientConfig):
    component = models.CharField(
        _("component"),
        max_length=50,
        choices=ComponentTypes.choices,
        default=ComponentTypes.zrc,
    )

    class Meta:
        verbose_name = _("Autorisatiecomponentconfiguratie")


class Applicatie(APIMixin, models.Model):
    """
    Client level of authorization
    """

    uuid = models.UUIDField(
        unique=True, default=uuid.uuid4, help_text="Unique resource identifier (UUID4)"
    )
    client_ids = ArrayField(
        models.CharField(max_length=50),
        verbose_name=_("client IDs"),
        help_text=_("Komma-gescheiden lijst van consumer identifiers (hun client_id)."),
    )
    label = models.CharField(
        max_length=100,
        help_text=_(
            "Een leesbare representatie van de applicatie, voor eindgebruikers."
        ),
    )
    heeft_alle_autorisaties = models.BooleanField(
        _("heeft alle autorisaties"),
        default=False,
        help_text=_(
            "Indien alle autorisaties gegeven zijn, dan hoeven deze "
            "niet individueel opgegeven te worden. Gebruik dit alleen "
            "als je de consumer helemaal vertrouwt."
        ),
    )

    def __str__(self):
        return f"Applicatie ({self.label})"


class Autorisatie(APIMixin, models.Model):
    applicatie = models.ForeignKey(
        "Applicatie",
        on_delete=models.CASCADE,
        related_name="autorisaties",
        verbose_name=_("applicatie"),
    )
    component = models.CharField(
        _("component"),
        max_length=50,
        choices=ComponentTypes.choices,
        help_text=_("Component waarop autorisatie van toepassing is."),
    )
    scopes = ArrayField(
        models.CharField(max_length=100),
        verbose_name=_("scopes"),
        help_text=_("Komma-gescheiden lijst van scope labels."),
    )

    # ZRC exclusive
    zaaktype = models.URLField(
        _("zaaktype"),
        help_text=_("URL naar het zaaktype waarop de autorisatie van toepassing is."),
        max_length=1000,
        blank=True,
    )

    # DRC exclusive
    informatieobjecttype = models.URLField(
        _("informatieobjecttype"),
        help_text=_(
            "URL naar het informatieobjecttype waarop de autorisatie van toepassing is."
        ),
        max_length=1000,
        blank=True,
    )

    # BRC exclusive
    besluittype = models.URLField(
        _("besluittype"),
        help_text=_(
            "URL naar het besluittype waarop de autorisatie van toepassing is."
        ),
        max_length=1000,
        blank=True,
    )

    # ZRC & DRC exclusive
    max_vertrouwelijkheidaanduiding = VertrouwelijkheidsAanduidingField(
        help_text=_("Maximaal toegelaten vertrouwelijkheidaanduiding (inclusief)."),
        blank=True,
    )

    def satisfy_vertrouwelijkheid(self, vertrouwelijkheidaanduiding: str) -> bool:
        max_confid_level = VertrouwelijkheidsAanduiding.get_choice(
            self.max_vertrouwelijkheidaanduiding
        ).order
        provided_confid_level = VertrouwelijkheidsAanduiding.get_choice(
            vertrouwelijkheidaanduiding
        ).order
        return max_confid_level >= provided_confid_level



from drf_yasg.app_settings import swagger_settings

from vng_api_common.notifications.api.views import NotificationView as _NotificationView


class NotificationView(_NotificationView):
    """
    Ontvang notificaties via webhook
    """

    swagger_schema = swagger_settings.DEFAULT_AUTO_SCHEMA_CLASS



from rest_framework.generics import CreateAPIView

from ..models import JWTSecret
from ..scopes import Scope
from .permissions import JWTCreatePermission
from .serializers import JWTSecretSerializer


class CreateJWTSecretView(CreateAPIView):
    swagger_schema = None

    model = JWTSecret
    serializer_class = JWTSecretSerializer
    permission_classes = (JWTCreatePermission,)
    required_scopes = {
        "create": Scope("autorisaties.credentials-registreren", private=True)
    }



"""
Utility module for Open API Specification 3.0.x.

This should get merged into gemma-zds-client, but some heavy refactoring is
needed for that.
"""
from typing import Union

import requests
import yaml
from drf_yasg import openapi

TYPE_MAP = {
    openapi.TYPE_OBJECT: dict,
    openapi.TYPE_STRING: str,
    openapi.TYPE_NUMBER: (float, int),
    openapi.TYPE_INTEGER: int,
    openapi.TYPE_BOOLEAN: bool,
    openapi.TYPE_ARRAY: list,
}


class SchemaFetcher:
    def __init__(self):
        self.cache = {}

    def fetch(self, url: str):
        """
        Fetch a YAML-based OAS 3.0.x schema.
        """
        if url in self.cache:
            return self.cache[url]

        response = requests.get(url)
        response.raise_for_status()

        spec = yaml.safe_load(response.content)
        spec_version = response.headers.get(
            "X-OAS-Version", spec.get("openapi", spec.get("swagger", ""))
        )
        if not spec_version.startswith("3.0"):
            raise ValueError("Unsupported spec version: {}".format(spec_version))

        self.cache[url] = spec
        return spec


def obj_has_shape(obj: Union[list, dict], schema: dict, resource: str) -> bool:
    """
    Compare an instance of an object with the expected shape from an OAS 3 schema.

    ..todo:: doesn't handle references and nested schema's yet.

    :param obj: the value retrieved from the endpoint, json-decoded to a dict or list
    :param schema: the OAS 3.0.x schema to test against, yaml-decoded to dict
    :param resource: the name of the resource to test the schape against
    """
    obj_schema = schema["components"]["schemas"][resource]

    required = obj_schema.get("required", [])
    for prop, prop_schema in obj_schema["properties"].items():
        if prop in required and prop not in obj:
            # missing required prop -> can't match the schema
            return False

        # can't compare something that isn't there...
        if prop not in obj:
            continue

        value = obj[prop]

        # TODO Handling references not yet implemented
        if "$ref" in prop_schema:
            continue

        # Allow None if property is nullable
        if value is None:
            if prop_schema.get("nullable", False):
                continue
            else:
                return False

        expected_type = TYPE_MAP[prop_schema["type"]]

        # type mismatch -> not what we're looking for
        if not isinstance(value, expected_type):
            return False

    return True


fetcher = SchemaFetcher()



# Generated by Django 2.0.13 on 2019-05-17 09:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("vng_api_common", "0003_auto_20190417_1145")]

    operations = [
        migrations.AddField(
            model_name="apicredential",
            name="user_id",
            field=models.CharField(default="", max_length=255, verbose_name="user id"),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="apicredential",
            name="user_representation",
            field=models.CharField(
                default="", max_length=255, verbose_name="user representation"
            ),
        ),
    ]



from unittest import skipUnless
from unittest.mock import patch

from django.apps import apps
from django.test import override_settings
from django.urls import reverse

from rest_framework import status
from rest_framework.test import APITestCase

from vng_api_common.tests import JWTAuthMixin

from ..constants import SCOPE_NOTIFICATIES_CONSUMEREN_LABEL
from ..models import NotificationsConfig, Subscription


@skipUnless(apps.is_installed("notifications.api"), "NRC API not enabled")
@override_settings(ROOT_URLCONF="vng_api_common.notifications.tests.urls")
class SubscriptionTests(JWTAuthMixin, APITestCase):

    scopes = [SCOPE_NOTIFICATIES_CONSUMEREN_LABEL]

    @classmethod
    def setUpTestData(cls):
        from notifications.datamodel.models import Kanaal

        super().setUpTestData()

        Kanaal.objects.create(naam="zaken")
        cls.config = NotificationsConfig.objects.create(
            api_root="http://testserver/api/"
        )

    @patch(
        "vng_api_common.notifications.models.Client.create",
        return_value={"url": "http://example.com/"},
    )
    def test_register_subscription(self, mock_create):
        url = reverse("abonnement-list", kwargs={"version": 1})
        sub = Subscription.objects.create(
            config=self.config,
            callback_url="http://testserver.com/api/callbacks",
            client_id="test",
            secret="test",
            channels=["zaken"],
        )

        sub.register()

        data = mock_create.call_args[1]["data"]

        # assert that we send valid data to the api

        response = self.client.post(url, data)

        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)



# Generated by Django 2.1.8 on 2019-05-02 04:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("authorizations", "0001_initial")]

    operations = [
        migrations.CreateModel(
            name="AuthorizationsConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("api_root", models.URLField(unique=True, verbose_name="api root")),
                (
                    "component",
                    models.CharField(
                        choices=[
                            ("ZRC", "ZRC"),
                            ("ZTC", "ZTC"),
                            ("DRC", "DRC"),
                            ("BRC", "BRC"),
                            ("NC", "NC"),
                            ("AC", "AC"),
                        ],
                        default="ZRC",
                        max_length=50,
                        verbose_name="component",
                    ),
                ),
            ],
            options={"verbose_name": "Autorisatiecomponentconfiguratie"},
        )
    ]



__all__ = [
    "API_VERSION",
    "BASE_REST_FRAMEWORK",
    "BASE_SWAGGER_SETTINGS",
    "LINK_FETCHER",
    "ZDS_CLIENT_CLASS",
    "GEMMA_URL_TEMPLATE",
    "GEMMA_URL_COMPONENTTYPE",
    "GEMMA_URL_INFORMATIEMODEL",
    "GEMMA_URL_INFORMATIEMODEL_VERSIE",
    "REDOC_SETTINGS",
    "NOTIFICATIONS_KANAAL",
    "NOTIFICATIONS_DISABLED",
]

API_VERSION = "1.0.0-rc1"  # semantic version

BASE_REST_FRAMEWORK = {
    "DEFAULT_RENDERER_CLASSES": (
        "djangorestframework_camel_case.render.CamelCaseJSONRenderer",
    ),
    "DEFAULT_PARSER_CLASSES": (
        "djangorestframework_camel_case.parser.CamelCaseJSONParser",
    ),
    "DEFAULT_AUTHENTICATION_CLASSES": (
        # 'oauth2_provider.contrib.rest_framework.OAuth2Authentication',
        # 'rest_framework.authentication.SessionAuthentication',
        # 'rest_framework.authentication.BasicAuthentication'
    ),
    # 'DEFAULT_PERMISSION_CLASSES': (
    #     'oauth2_provider.contrib.rest_framework.TokenHasReadWriteScope',
    #     # 'rest_framework.permissions.IsAuthenticated',
    #     # 'rest_framework.permissions.AllowAny',
    # ),
    "DEFAULT_VERSIONING_CLASS": "rest_framework.versioning.URLPathVersioning",
    #
    # # Generic view behavior
    # 'DEFAULT_PAGINATION_CLASS': 'ztc.api.utils.pagination.HALPagination',
    "DEFAULT_FILTER_BACKENDS": (
        "vng_api_common.filters.Backend",
        # 'rest_framework.filters.SearchFilter',
        # 'rest_framework.filters.OrderingFilter',
    ),
    #
    # # Filtering
    # 'SEARCH_PARAM': 'zoek',  # 'search',
    "ORDERING_PARAM": "ordering",  # 'ordering',
    #
    # Versioning
    "DEFAULT_VERSION": "1",  # NOT to be confused with API_VERSION - it's the major version part
    "ALLOWED_VERSIONS": ("1",),
    "VERSION_PARAM": "version",
    #
    # # Exception handling
    "EXCEPTION_HANDLER": "vng_api_common.views.exception_handler",
    "TEST_REQUEST_DEFAULT_FORMAT": "json",
}

BASE_SWAGGER_SETTINGS = {
    # 'SECURITY_DEFINITIONS': {
    #     'OAuth2': {
    #         'type': 'oauth2',
    #         'flow': 'application',
    #         'tokenUrl': '/oauth2/token/',
    #         'scopes': {
    #             'write': 'Schrijftoegang tot de catalogus en gerelateerde objecten.',
    #             'read': 'Leestoegang tot de catalogus en gerelateerde objecten.'
    #         }
    #     },
    #     'Bearer': {
    #         'type': 'apiKey',
    #         'name': 'Authorization',
    #         'in': 'header'
    #     },
    # },
    "DEFAULT_AUTO_SCHEMA_CLASS": "vng_api_common.inspectors.view.AutoSchema",
    "DEFAULT_INFO": "must.be.overridden",
    "DEFAULT_FIELD_INSPECTORS": (
        # GeometryFieldInspector has external dependencies, and is opt-in
        # 'vng_api_common.inspectors.geojson.GeometryFieldInspector',
        "vng_api_common.inspectors.fields.HyperlinkedIdentityFieldInspector",
        "vng_api_common.inspectors.fields.ReadOnlyFieldInspector",
        "vng_api_common.inspectors.fields.JSONFieldInspector",
        "vng_api_common.inspectors.polymorphic.PolymorphicSerializerInspector",
        "vng_api_common.inspectors.fields.GegevensGroepInspector",
        "drf_yasg.inspectors.CamelCaseJSONFilter",
        "drf_yasg.inspectors.RecursiveFieldInspector",
        "drf_yasg.inspectors.ReferencingSerializerInspector",
        "drf_yasg.inspectors.ChoiceFieldInspector",
        "drf_yasg.inspectors.FileFieldInspector",
        "drf_yasg.inspectors.DictFieldInspector",
        "drf_yasg.inspectors.HiddenFieldInspector",
        "drf_yasg.inspectors.RelatedFieldInspector",
        "drf_yasg.inspectors.SimpleFieldInspector",
        "drf_yasg.inspectors.StringDefaultFieldInspector",
    ),
    "DEFAULT_FILTER_INSPECTORS": ("vng_api_common.inspectors.query.FilterInspector",),
}

REDOC_SETTINGS = {"EXPAND_RESPONSES": "200,201", "SPEC_URL": "openapi.json"}

# See: https://github.com/Rebilly/ReDoc#redoc-options-object
LINK_FETCHER = "requests.get"

ZDS_CLIENT_CLASS = "zds_client.Client"

GEMMA_URL_TEMPLATE = "https://www.gemmaonline.nl/index.php/{informatiemodel}_{versie}/doc/{componenttype}/{component}"
GEMMA_URL_COMPONENTTYPE = "objecttype"
GEMMA_URL_INFORMATIEMODEL = "Rgbz"
GEMMA_URL_INFORMATIEMODEL_VERSIE = "2.0"

# notifications configuration

NOTIFICATIONS_KANAAL = None
NOTIFICATIONS_DISABLED = False
